from datetime import datetime, timedelta, timezone

from flask import jsonify
from mlflow.server.handlers import _get_model_registry_store, _get_tracking_store, catch_mlflow_exception

from mlflow_oidc_auth.permissions import NO_PERMISSIONS
from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.user import create_user, generate_token
from mlflow_oidc_auth.utils import (
    effective_experiment_permission,
    effective_prompt_permission,
    effective_registered_model_permission,
    fetch_all_registered_models,
    fetch_all_prompts,
    get_is_admin,
    get_optional_request_param,
    get_request_param,
    get_username,
)


@catch_mlflow_exception
def create_new_user():
    username = get_request_param("username")
    display_name = get_request_param("display_name")
    is_admin = bool(get_optional_request_param("is_admin") or False)
    is_service_account = bool(get_optional_request_param("is_service_account") or False)
    status, message = create_user(username, display_name, is_admin, is_service_account)
    if status:
        return (jsonify({"message": message}), 201)
    else:
        return (jsonify({"message": message}), 200)


@catch_mlflow_exception
def get_user():
    username = get_request_param("username")
    user = store.get_user(username)
    return jsonify({"user": user.to_json()})


@catch_mlflow_exception
def delete_user():
    username = get_request_param("username")
    store.delete_user(username)
    return jsonify({"message": f"Account {username} has been deleted"})


@catch_mlflow_exception
def create_user_access_token():
    username = get_request_param("username")
    expiration_str = get_request_param("expiration")
    # Handle ISO 8601 with 'Z' (UTC) at the end
    if expiration_str:
        if expiration_str.endswith("Z"):
            expiration_str = expiration_str[:-1] + "+00:00"
        expiration = datetime.fromisoformat(expiration_str)
        now = datetime.now(timezone.utc)
        if expiration < now:
            return jsonify({"message": "Expiration date must be in the future"}), 400
        if expiration > now + timedelta(days=366):
            return jsonify({"message": "Expiration date must be less than 1 year in the future"}), 400
    else:
        expiration = None
    user = store.get_user(username)
    if user is None:
        return jsonify({"message": f"User {username} not found"}), 404
    new_token = generate_token()
    store.update_user(username=username, password=new_token, password_expiration=expiration)
    return jsonify({"token": new_token, "message": f"Token for {username} has been created"})


@catch_mlflow_exception
def update_username_password():
    new_password = generate_token()
    store.update_user(username=get_username(), password=new_password)
    return jsonify({"token": new_password})


# TODO: move filtering logic to store
@catch_mlflow_exception
def list_user_experiments(username):
    current_user = store.get_user(get_username())
    all_experiments = _get_tracking_store().search_experiments()
    is_admin = get_is_admin()

    if is_admin:
        list_experiments = all_experiments
    else:
        if username == current_user.username:
            list_experiments = [
                exp for exp in all_experiments if effective_experiment_permission(exp.experiment_id, username).permission.name != NO_PERMISSIONS.name
            ]
        else:
            list_experiments = [
                exp for exp in all_experiments if effective_experiment_permission(exp.experiment_id, current_user.username).permission.can_manage
            ]

    experiments_list = [
        {
            "name": _get_tracking_store().get_experiment(exp.experiment_id).name,
            "id": exp.experiment_id,
            "permission": (perm := effective_experiment_permission(exp.experiment_id, username)).permission.name,
            "type": perm.type,
        }
        for exp in list_experiments
    ]
    return experiments_list


@catch_mlflow_exception
def list_user_models(username):
    all_registered_models = fetch_all_registered_models()
    current_user = store.get_user(get_username())
    is_admin = get_is_admin()
    if is_admin:
        list_registered_models = all_registered_models
    else:
        if username == current_user.username:
            list_registered_models = [
                model for model in all_registered_models if effective_registered_model_permission(model.name, username).permission.name != NO_PERMISSIONS.name
            ]
        else:
            list_registered_models = [
                model for model in all_registered_models if effective_registered_model_permission(model.name, current_user.username).permission.can_manage
            ]
    models = [
        {
            "name": model.name,
            "permission": (perm := effective_registered_model_permission(model.name, username)).permission.name,
            "type": perm.type,
        }
        for model in list_registered_models
    ]
    return models


@catch_mlflow_exception
def list_user_prompts(username):
    all_registered_models = fetch_all_prompts()
    current_user = store.get_user(get_username())
    is_admin = get_is_admin()
    if is_admin:
        list_registered_models = all_registered_models
    else:
        if username == current_user.username:
            list_registered_models = [
                model for model in all_registered_models if effective_prompt_permission(model.name, username).permission.name != NO_PERMISSIONS.name
            ]
        else:
            list_registered_models = [
                model for model in all_registered_models if effective_prompt_permission(model.name, current_user.username).permission.can_manage
            ]
    models = [
        {
            "name": model.name,
            "permission": (perm := effective_prompt_permission(model.name, username)).permission.name,
            "type": perm.type,
        }
        for model in list_registered_models
    ]
    return models


# TODO: use to_json
@catch_mlflow_exception
def list_users():
    service_account = bool(get_optional_request_param("service") or False)
    # is_admin = get_is_admin()
    # if is_admin:
    #     users = [user.username for user in store.list_users()]
    # else:
    #     users = [get_username()]
    users = [user.username for user in store.list_users(is_service_account=service_account)]
    return users


@catch_mlflow_exception
def update_user_admin():
    is_admin = get_request_param("is_admin").strip().lower() == "true" if get_request_param("is_admin") else False
    store.update_user(username=get_username(), is_admin=is_admin)
    return jsonify({"is_admin": is_admin})


@catch_mlflow_exception
def get_current_user():
    return store.get_user(get_username()).to_json()



"""
Unified logging module for MLflow OIDC Auth Plugin.

This module provides a centralized logging solution that works consistently
across both Flask and FastAPI modes. It automatically detects the server mode
and configures appropriate loggers.
"""

import logging
import os
import sys
from typing import Optional


class UnifiedLogger:
    """
    Unified logger that works seamlessly in both Flask and FastAPI modes.

    This class automatically detects the server mode and provides a consistent
    logging interface across the application. It supports configuration through
    environment variables and provides proper formatting for different server types.
    """

    _instance: Optional["UnifiedLogger"] = None
    _logger: Optional[logging.Logger] = None
    _server_mode: Optional[str] = None

    def __new__(cls) -> "UnifiedLogger":
        """Singleton pattern to ensure only one logger instance."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        """Initialize the unified logger with automatic server mode detection."""
        if self._logger is None:
            self._detect_server_mode()
            self._setup_logger()

    def _detect_server_mode(self) -> None:
        """
        Detect the server mode (Flask or FastAPI) based on available modules.

        This method checks the current runtime environment to determine whether
        the application is running in Flask mode (MLflow server) or FastAPI mode.
        """
        try:
            # Check if we're in a FastAPI context
            import uvicorn

            if "uvicorn" in sys.modules or "fastapi" in sys.modules:
                self._server_mode = "fastapi"
            else:
                self._server_mode = "flask"
        except ImportError:
            # Fall back to Flask mode if FastAPI modules are not available
            self._server_mode = "flask"

    def _setup_logger(self) -> None:
        """
        Set up the logger based on the detected server mode.

        This method configures the appropriate logger for the current server mode:
        - Flask mode: Uses mlflow.server.app logger or creates a new one
        - FastAPI mode: Uses uvicorn logger or creates a compatible one
        """
        log_level = os.environ.get("LOG_LEVEL", "INFO").upper()

        if self._server_mode == "fastapi":
            # For FastAPI mode, use uvicorn logger or create a compatible one
            try:
                self._logger = logging.getLogger("uvicorn")
                if not self._logger.handlers:
                    # If uvicorn logger doesn't have handlers, set up our own
                    self._setup_custom_logger("mlflow_oidc_auth.fastapi", log_level)
            except Exception:
                # Fallback to custom logger if uvicorn logger is not available
                self._setup_custom_logger("mlflow_oidc_auth.fastapi", log_level)
        else:
            # For Flask mode, try to use app.logger or create a custom logger
            try:
                from mlflow.server import app

                self._logger = app.logger
                # Ensure the logger level is set according to environment
                self._logger.setLevel(getattr(logging, log_level, logging.INFO))
            except (ImportError, AttributeError):
                # Fallback to custom logger if Flask app is not available
                self._setup_custom_logger("mlflow_oidc_auth.flask", log_level)

    def _setup_custom_logger(self, logger_name: str, log_level: str) -> None:
        """
        Set up a custom logger with proper formatting.

        Args:
            logger_name: Name for the logger
            log_level: Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        """
        self._logger = logging.getLogger(logger_name)
        self._logger.setLevel(getattr(logging, log_level, logging.INFO))

        # Only add handler if logger doesn't have any to avoid duplicates
        if not self._logger.handlers:
            # Create console handler with formatting
            handler = logging.StreamHandler(sys.stdout)
            handler.setLevel(getattr(logging, log_level, logging.INFO))

            # Create formatter
            formatter = logging.Formatter("[%(asctime)s] %(levelname)s in %(name)s (%(filename)s:%(lineno)d): %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
            handler.setFormatter(formatter)

            self._logger.addHandler(handler)

    def get_logger(self) -> logging.Logger:
        """
        Get the configured logger instance.

        Returns:
            logging.Logger: The configured logger instance for the current server mode
        """
        if self._logger is None:
            self._setup_logger()
        # Type assertion is safe here as _setup_logger always sets _logger
        assert self._logger is not None, "Logger should be initialized"
        return self._logger

    def get_server_mode(self) -> str:
        """
        Get the detected server mode.

        Returns:
            str: The detected server mode ('flask' or 'fastapi')
        """
        return self._server_mode or "unknown"

    # Convenience methods for direct logging
    def debug(self, message: str, *args, **kwargs) -> None:
        """Log a debug message."""
        self.get_logger().debug(message, *args, **kwargs)

    def info(self, message: str, *args, **kwargs) -> None:
        """Log an info message."""
        self.get_logger().info(message, *args, **kwargs)

    def warning(self, message: str, *args, **kwargs) -> None:
        """Log a warning message."""
        self.get_logger().warning(message, *args, **kwargs)

    def error(self, message: str, *args, **kwargs) -> None:
        """Log an error message."""
        self.get_logger().error(message, *args, **kwargs)

    def critical(self, message: str, *args, **kwargs) -> None:
        """Log a critical message."""
        self.get_logger().critical(message, *args, **kwargs)


# Create the global logger instance
_unified_logger = UnifiedLogger()


# Export convenience functions for easy import and use
def get_logger() -> logging.Logger:
    """
    Get the unified logger instance.

    This function provides easy access to the configured logger that works
    in both Flask and FastAPI modes. Use this in your modules instead of
    creating separate loggers.

    Returns:
        logging.Logger: The configured logger instance

    Example:
        from mlflow_oidc_auth.logger import get_logger
        logger = get_logger()
        logger.info("This works in both Flask and FastAPI modes")
    """
    return _unified_logger.get_logger()


def get_server_mode() -> str:
    """
    Get the detected server mode.

    Returns:
        str: The detected server mode ('flask' or 'fastapi')
    """
    return _unified_logger.get_server_mode()


# Export convenience logging functions
def debug(message: str, *args, **kwargs) -> None:
    """Log a debug message using the unified logger."""
    _unified_logger.debug(message, *args, **kwargs)


def info(message: str, *args, **kwargs) -> None:
    """Log an info message using the unified logger."""
    _unified_logger.info(message, *args, **kwargs)


def warning(message: str, *args, **kwargs) -> None:
    """Log a warning message using the unified logger."""
    _unified_logger.warning(message, *args, **kwargs)


def error(message: str, *args, **kwargs) -> None:
    """Log an error message using the unified logger."""
    _unified_logger.error(message, *args, **kwargs)


def critical(message: str, *args, **kwargs) -> None:
    """Log a critical message using the unified logger."""
    _unified_logger.critical(message, *args, **kwargs)



"""
Test cases for mlflow_oidc_auth.utils.decorators module.

This module contains comprehensive tests for permission decorators
that control access to MLflow operations.
"""

import unittest
from unittest.mock import patch

from flask import Flask

from mlflow_oidc_auth.utils import (
    check_experiment_permission,
    check_registered_model_permission,
    check_prompt_permission,
    check_admin_permission,
)


class TestDecorators(unittest.TestCase):
    """Test cases for decorator utility functions."""

    def setUp(self) -> None:
        """Set up test environment with Flask application context."""
        self.app = Flask(__name__)
        self.app.config["TESTING"] = True
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()

    def tearDown(self) -> None:
        """Clean up test environment."""
        self.app_context.pop()

    @patch("mlflow_oidc_auth.utils.decorators.store")
    @patch("mlflow_oidc_auth.utils.decorators.get_is_admin")
    @patch("mlflow_oidc_auth.utils.decorators.get_username")
    @patch("mlflow_oidc_auth.utils.decorators.get_experiment_id")
    @patch("mlflow_oidc_auth.utils.decorators.can_manage_experiment")
    @patch("mlflow_oidc_auth.utils.decorators.make_forbidden_response")
    def test_check_experiment_permission(
        self,
        mock_make_forbidden_response,
        mock_can_manage_experiment,
        mock_get_experiment_id,
        mock_get_username,
        mock_get_is_admin,
        mock_store,
    ):
        """Test experiment permission decorator functionality."""
        with self.app.test_request_context():
            mock_get_is_admin.return_value = False
            mock_get_username.return_value = "user"
            mock_get_experiment_id.return_value = "exp_id"
            mock_can_manage_experiment.return_value = False
            mock_make_forbidden_response.return_value = "forbidden"

            @check_experiment_permission
            def mock_func():
                return "success"

            self.assertEqual(mock_func(), "forbidden")

            mock_can_manage_experiment.return_value = True
            self.assertEqual(mock_func(), "success")

            # Admin always allowed
            mock_get_is_admin.return_value = True
            self.assertEqual(mock_func(), "success")

    @patch("mlflow_oidc_auth.utils.decorators.store")
    @patch("mlflow_oidc_auth.utils.decorators.get_is_admin")
    @patch("mlflow_oidc_auth.utils.decorators.get_username")
    @patch("mlflow_oidc_auth.utils.decorators.get_model_name")
    @patch("mlflow_oidc_auth.utils.decorators.can_manage_registered_model")
    @patch("mlflow_oidc_auth.utils.decorators.make_forbidden_response")
    def test_check_registered_model_permission(
        self,
        mock_make_forbidden_response,
        mock_can_manage_registered_model,
        mock_get_model_name,
        mock_get_username,
        mock_get_is_admin,
        mock_store,
    ):
        """Test registered model permission decorator functionality."""
        with self.app.test_request_context():
            mock_get_is_admin.return_value = False
            mock_get_username.return_value = "user"
            mock_get_model_name.return_value = "model_name"
            mock_can_manage_registered_model.return_value = False
            mock_make_forbidden_response.return_value = "forbidden"

            @check_registered_model_permission
            def mock_func():
                return "success"

            self.assertEqual(mock_func(), "forbidden")

            mock_can_manage_registered_model.return_value = True
            self.assertEqual(mock_func(), "success")

            # Admin always allowed
            mock_get_is_admin.return_value = True
            self.assertEqual(mock_func(), "success")

    @patch("mlflow_oidc_auth.utils.decorators.store")
    @patch("mlflow_oidc_auth.utils.decorators.get_is_admin")
    @patch("mlflow_oidc_auth.utils.decorators.get_username")
    @patch("mlflow_oidc_auth.utils.decorators.get_model_name")
    @patch("mlflow_oidc_auth.utils.decorators.can_manage_registered_model")
    @patch("mlflow_oidc_auth.utils.decorators.make_forbidden_response")
    def test_check_prompt_permission(
        self,
        mock_make_forbidden_response,
        mock_can_manage_registered_model,
        mock_get_model_name,
        mock_get_username,
        mock_get_is_admin,
        mock_store,
    ):
        """Test prompt permission decorator functionality."""
        with self.app.test_request_context():
            mock_get_is_admin.return_value = False
            mock_get_username.return_value = "user"
            mock_get_model_name.return_value = "prompt_name"
            mock_can_manage_registered_model.return_value = False
            mock_make_forbidden_response.return_value = "forbidden"

            @check_prompt_permission
            def mock_func():
                return "success"

            self.assertEqual(mock_func(), "forbidden")

            mock_can_manage_registered_model.return_value = True
            self.assertEqual(mock_func(), "success")

            # Admin always allowed
            mock_get_is_admin.return_value = True
            self.assertEqual(mock_func(), "success")

    @patch("mlflow_oidc_auth.utils.decorators.store")
    @patch("mlflow_oidc_auth.utils.decorators.get_username")
    @patch("mlflow_oidc_auth.utils.decorators.get_is_admin")
    @patch("mlflow_oidc_auth.utils.decorators.make_forbidden_response")
    def test_check_admin_permission(self, mock_make_forbidden_response, mock_get_is_admin, mock_get_username, mock_store):
        """Test admin permission decorator functionality."""
        with self.app.test_request_context():
            mock_get_username.return_value = "user"
            mock_get_is_admin.return_value = False
            mock_make_forbidden_response.return_value = "forbidden"

            @check_admin_permission
            def mock_func():
                return "success"

            self.assertEqual(mock_func(), "forbidden")

            # Admin allowed
            mock_get_is_admin.return_value = True
            self.assertEqual(mock_func(), "success")


if __name__ == "__main__":
    unittest.main()



"""
Type definitions for MLflow OIDC Auth utilities.

This module provides common type definitions used across the MLflow OIDC Auth system.
"""

from typing import NamedTuple
from mlflow_oidc_auth.permissions import Permission


class PermissionResult(NamedTuple):
    """
    Result object containing permission information and its source.

    This class encapsulates both the permission details and metadata about
    where the permission was determined from (e.g., user, group, regex, fallback).

    Attributes:
        permission (Permission): The Permission object containing access rights
        type (str): String indicating the source type (e.g., 'user', 'group', 'regex', 'fallback')
    """

    permission: Permission
    type: str



"""
Configuration endpoint for dynamic runtime configuration.

This module provides endpoints to expose runtime configuration
to the frontend, including proxy path information.
"""

from flask import jsonify, request, Response
from mlflow_oidc_auth.routes import UI_ROOT


def get_runtime_config() -> Response:
    """
    Get runtime configuration for the frontend application.

    This endpoint provides configuration that may change at runtime,
    particularly proxy path information. With ProxyFix middleware configured,
    Flask's request object automatically contains the correct values.

    Returns:
        Response: JSON response containing runtime configuration:
            - basePath: The base path for the application
            - uiPath: The relative path where UI files are served
    """

    config = {"basePath": request.script_root, "uiPath": request.script_root + UI_ROOT}

    return jsonify(config)



from typing import Optional

import requests
from authlib.integrations.flask_client import OAuth
from authlib.jose import jwt
from authlib.jose.errors import BadSignatureError
from flask import request
from mlflow.server import app

from mlflow_oidc_auth.config import config
from mlflow_oidc_auth.logger import get_logger
from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.user import create_user, populate_groups, update_user

logger = get_logger()

_oauth_instance: Optional[OAuth] = None


def get_oauth_instance(app) -> OAuth:
    # returns a singleton instance of OAuth
    # to avoid circular imports
    global _oauth_instance

    if _oauth_instance is None:
        _oauth_instance = OAuth(app)
        _oauth_instance.register(
            name="oidc",
            client_id=config.OIDC_CLIENT_ID,
            client_secret=config.OIDC_CLIENT_SECRET,
            server_metadata_url=config.OIDC_DISCOVERY_URL,
            client_kwargs={"scope": config.OIDC_SCOPE},
        )
    return _oauth_instance


def _get_oidc_jwks(clear_cache: bool = False):
    from mlflow_oidc_auth.app import cache

    if clear_cache:
        logger.debug("Clearing JWKS cache")
        cache.delete("jwks")
    jwks = cache.get("jwks")
    if jwks:
        logger.debug("JWKS cache hit")
        return jwks
    logger.debug("JWKS cache miss")
    if config.OIDC_DISCOVERY_URL is None:
        raise ValueError("OIDC_DISCOVERY_URL is not set in the configuration")
    metadata = requests.get(config.OIDC_DISCOVERY_URL).json()
    jwks_uri = metadata.get("jwks_uri")
    jwks = requests.get(jwks_uri).json()
    cache.set("jwks", jwks, timeout=3600)
    return jwks


def validate_token(token):
    try:
        jwks = _get_oidc_jwks()
        payload = jwt.decode(token, jwks)
        payload.validate()
        return payload
    except BadSignatureError as e:
        logger.warning("Token validation failed. Attempting JWKS refresh. Error: %s", str(e))
        jwks = _get_oidc_jwks(clear_cache=True)
        try:
            payload = jwt.decode(token, jwks)
            payload.validate()
            return payload
        except BadSignatureError as e:
            logger.error("Token validation failed after JWKS refresh. Error: %s", str(e))
            raise
        except Exception as e:
            logger.error("Unexpected error during token validation: %s", str(e))
            raise


def authenticate_request_basic_auth() -> bool:
    if request.authorization is None:
        return False
    username = request.authorization.username
    password = request.authorization.password
    logger.debug("Authenticating user %s", username)
    if username is not None and password is not None and store.authenticate_user(username.lower(), password):
        logger.debug("User %s authenticated", username)
        return True
    else:
        logger.debug("User %s not authenticated", username)
        return False


def authenticate_request_bearer_token() -> bool:
    if request.authorization and request.authorization.token:
        token = request.authorization.token
        try:
            user = validate_token(token)
            logger.debug("User %s authenticated", user.get("email"))
            return True
        except Exception as e:
            logger.error(f"JWT auth failed: {str(e)}")
            return False
    else:
        logger.debug("No authorization token found")
        return False


def handle_token_validation(oauth_instance: OAuth):
    """Validate the token and handle JWKS refresh if necessary."""
    if getattr(oauth_instance, "oidc", None) is None:
        logger.error("OAuth instance or OIDC is not properly initialized")
        return None
    if oauth_instance.oidc is None or not hasattr(oauth_instance.oidc, "authorize_access_token") or not callable(oauth_instance.oidc.authorize_access_token):
        logger.error("OIDC client is not properly initialized or missing 'authorize_access_token' method")
        return None
    try:
        token = oauth_instance.oidc.authorize_access_token()
    except BadSignatureError:
        logger.warning("Bad signature detected. Refreshing JWKS keys.")
        if not hasattr(oauth_instance.oidc, "load_server_metadata") or not callable(oauth_instance.oidc.load_server_metadata):
            logger.error("OIDC client is missing 'load_server_metadata' method")
            return None
        oauth_instance.oidc.load_server_metadata()
        try:
            token = oauth_instance.oidc.authorize_access_token()
        except BadSignatureError:
            logger.error("Bad signature persists after JWKS refresh. Token verification failed.")
            return None
    logger.debug(f"Token: {token}")
    return token


def handle_user_and_group_management(token) -> list[str]:
    """Handle user and group management based on the token. Returns list of error messages or empty list."""
    errors = []
    email = token["userinfo"].get("email") or token["userinfo"].get("preferred_username")
    display_name = token["userinfo"].get("name")
    if not email:
        errors.append("User profile error: No email provided in OIDC userinfo.")
    if not display_name:
        errors.append("User profile error: No display name provided in OIDC userinfo.")
    if errors:
        return errors

    # Get user groups
    try:
        if config.OIDC_GROUP_DETECTION_PLUGIN:
            import importlib

            user_groups = importlib.import_module(config.OIDC_GROUP_DETECTION_PLUGIN).get_user_groups(token["access_token"])
        else:
            user_groups = token["userinfo"][config.OIDC_GROUPS_ATTRIBUTE]
    except Exception as e:
        logger.error(f"Group detection error: {str(e)}")
        errors.append("Group detection error: Failed to get user groups")
        return errors

    logger.debug(f"User groups: {user_groups}")

    is_admin = config.OIDC_ADMIN_GROUP_NAME in user_groups
    if not is_admin and not any(group in user_groups for group in config.OIDC_GROUP_NAME):
        errors.append("Authorization error: User is not allowed to login.")
        return errors

    try:
        create_user(username=email.lower(), display_name=display_name, is_admin=is_admin)
        populate_groups(group_names=user_groups)
        update_user(username=email.lower(), group_names=user_groups)
    except Exception as e:
        logger.error(f"User/group DB error: {str(e)}")
        errors.append("User/group DB error: Failed to update user/groups")

    return errors


def process_oidc_callback(request, session) -> tuple[Optional[str], list[str]]:
    """
    Process the OIDC callback logic.
    Returns (email, error_list) tuple.
    """
    import html

    errors = []

    # Handle OIDC error response
    error_param = request.args.get("error")
    error_description = request.args.get("error_description")
    if error_param:
        safe_desc = html.escape(error_description) if error_description else ""
        errors.append("OIDC provider error: An error occurred during the OIDC authentication process.")
        if safe_desc:
            errors.append(f"{safe_desc}")
        return None, errors

    # State check
    state = request.args.get("state")
    if "oauth_state" not in session:
        errors.append("Session error: Missing OAuth state in session. Please try logging in again.")
        return None, errors
    if state != session["oauth_state"]:
        errors.append("Security error: Invalid state parameter. Possible CSRF detected.")
        return None, errors

    oauth_instance = get_oauth_instance(app)
    if oauth_instance is None or getattr(oauth_instance, "oidc", None) is None:
        logger.error("OAuth instance or OIDC is not properly initialized")
        errors.append("Server error: OAuth instance or OIDC is not properly initialized. Please contact the administrator.")
        return None, errors

    token = handle_token_validation(oauth_instance)
    if token is None:
        errors.append("OIDC token error: Invalid token signature or token could not be validated.")
        return None, errors

    # User and group management
    user_errors = handle_user_and_group_management(token)
    if user_errors:
        errors.extend(user_errors)
        return None, errors

    userinfo = getattr(token, "userinfo", None)
    if userinfo is None and isinstance(token, dict):
        userinfo = token.get("userinfo")
    if not isinstance(userinfo, dict):
        errors.append("OIDC token error: 'userinfo' is missing or not a dictionary.")
        return None, errors
    email = userinfo.get("email") or userinfo.get("preferred_username")
    if email is None:
        errors.append("OIDC token error: 'email' is missing in userinfo.")
        return None, errors
    return email.lower(), []



import re
from typing import Any, Callable, Dict, Optional

from flask import Request, redirect, render_template, request, session, url_for
from mlflow.protos.model_registry_pb2 import (
    CreateModelVersion,
    DeleteModelVersion,
    DeleteModelVersionTag,
    DeleteRegisteredModel,
    DeleteRegisteredModelAlias,
    DeleteRegisteredModelTag,
    GetLatestVersions,
    GetModelVersion,
    GetModelVersionByAlias,
    GetModelVersionDownloadUri,
    GetRegisteredModel,
    RenameRegisteredModel,
    SetModelVersionTag,
    SetRegisteredModelAlias,
    SetRegisteredModelTag,
    TransitionModelVersionStage,
    UpdateModelVersion,
    UpdateRegisteredModel,
)
from mlflow.protos.service_pb2 import (
    CreateLoggedModel,
    CreateRun,
    DeleteExperiment,
    DeleteExperimentTag,
    DeleteLoggedModel,
    DeleteLoggedModelTag,
    DeleteRun,
    DeleteTag,
    FinalizeLoggedModel,
    GetExperiment,
    GetExperimentByName,
    GetLoggedModel,
    GetMetricHistory,
    GetRun,
    ListArtifacts,
    LogBatch,
    LogLoggedModelParamsRequest,
    LogMetric,
    LogModel,
    LogParam,
    RestoreExperiment,
    RestoreRun,
    SetExperimentTag,
    SetLoggedModelTags,
    SetTag,
    UpdateExperiment,
    UpdateRun,
)
from mlflow.server.handlers import get_endpoints
from mlflow.utils.rest_utils import _REST_API_PATH_PREFIX

import mlflow_oidc_auth.responses as responses
from mlflow_oidc_auth import routes
from mlflow_oidc_auth.auth import authenticate_request_basic_auth, authenticate_request_bearer_token
from mlflow_oidc_auth.config import config
from mlflow_oidc_auth.utils import get_is_admin
from mlflow_oidc_auth.validators import (
    validate_can_create_user,
    validate_can_delete_experiment,
    validate_can_delete_experiment_artifact_proxy,
    validate_can_delete_logged_model,
    validate_can_delete_registered_model,
    validate_can_delete_run,
    validate_can_delete_user,
    validate_can_get_user_token,
    validate_can_manage_experiment,
    validate_can_manage_registered_model,
    validate_can_read_experiment,
    validate_can_read_experiment_artifact_proxy,
    validate_can_read_experiment_by_name,
    validate_can_read_logged_model,
    validate_can_read_registered_model,
    validate_can_read_run,
    validate_can_update_experiment,
    validate_can_update_experiment_artifact_proxy,
    validate_can_update_logged_model,
    validate_can_update_registered_model,
    validate_can_update_run,
    validate_can_update_user_admin,
    validate_can_update_user_password,
)


def _is_unprotected_route(path: str) -> bool:
    return path.startswith(
        (
            "/health",
            "/login",
            "/callback",
            "/oidc/static",
            "/metrics",
        )
    )


BEFORE_REQUEST_HANDLERS = {
    # Routes for experiments
    ## CreateExperiment: _validate_can_manage_experiment,
    GetExperiment: validate_can_read_experiment,
    GetExperimentByName: validate_can_read_experiment_by_name,
    DeleteExperiment: validate_can_delete_experiment,
    RestoreExperiment: validate_can_delete_experiment,
    UpdateExperiment: validate_can_update_experiment,
    SetExperimentTag: validate_can_update_experiment,
    DeleteExperimentTag: validate_can_update_experiment,
    # # Routes for runs
    CreateRun: validate_can_update_experiment,
    GetRun: validate_can_read_run,
    DeleteRun: validate_can_delete_run,
    RestoreRun: validate_can_delete_run,
    UpdateRun: validate_can_update_run,
    LogMetric: validate_can_update_run,
    LogBatch: validate_can_update_run,
    LogModel: validate_can_update_run,
    SetTag: validate_can_update_run,
    DeleteTag: validate_can_update_run,
    LogParam: validate_can_update_run,
    GetMetricHistory: validate_can_read_run,
    ListArtifacts: validate_can_read_run,
    # # Routes for model registry
    GetRegisteredModel: validate_can_read_registered_model,
    DeleteRegisteredModel: validate_can_delete_registered_model,
    UpdateRegisteredModel: validate_can_update_registered_model,
    RenameRegisteredModel: validate_can_update_registered_model,
    GetLatestVersions: validate_can_read_registered_model,
    CreateModelVersion: validate_can_update_registered_model,
    GetModelVersion: validate_can_read_registered_model,
    DeleteModelVersion: validate_can_delete_registered_model,
    UpdateModelVersion: validate_can_update_registered_model,
    TransitionModelVersionStage: validate_can_update_registered_model,
    GetModelVersionDownloadUri: validate_can_read_registered_model,
    SetRegisteredModelTag: validate_can_update_registered_model,
    DeleteRegisteredModelTag: validate_can_update_registered_model,
    SetModelVersionTag: validate_can_update_registered_model,
    DeleteModelVersionTag: validate_can_delete_registered_model,
    SetRegisteredModelAlias: validate_can_update_registered_model,
    DeleteRegisteredModelAlias: validate_can_delete_registered_model,
    GetModelVersionByAlias: validate_can_read_registered_model,
}


def _get_before_request_handler(request_class):
    return BEFORE_REQUEST_HANDLERS.get(request_class)


BEFORE_REQUEST_VALIDATORS = {(http_path, method): handler for http_path, handler, methods in get_endpoints(_get_before_request_handler) for method in methods}

BEFORE_REQUEST_VALIDATORS.update(
    {
        (routes.CREATE_ACCESS_TOKEN, "PATCH"): validate_can_get_user_token,
        # (SIGNUP, "GET"): validate_can_create_user,
        # (routes.GET_USER, "GET"): validate_can_read_user,
        (routes.CREATE_USER, "POST"): validate_can_create_user,
        (routes.UPDATE_USER_PASSWORD, "PATCH"): validate_can_update_user_password,
        (routes.UPDATE_USER_ADMIN, "PATCH"): validate_can_update_user_admin,
        (routes.DELETE_USER, "DELETE"): validate_can_delete_user,
        (routes.USER_EXPERIMENT_PERMISSIONS, "GET"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PERMISSIONS, "POST"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PERMISSION_DETAIL, "GET"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PERMISSION_DETAIL, "POST"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PERMISSION_DETAIL, "PATCH"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PERMISSION_DETAIL, "DELETE"): validate_can_manage_experiment,
        (routes.EXPERIMENT_USER_PERMISSIONS, "GET"): validate_can_manage_experiment,
        (routes.EXPERIMENT_USER_PERMISSIONS, "POST"): validate_can_manage_experiment,
        (routes.EXPERIMENT_USER_PERMISSION_DETAIL, "GET"): validate_can_manage_experiment,
        (routes.EXPERIMENT_USER_PERMISSION_DETAIL, "POST"): validate_can_manage_experiment,
        (routes.EXPERIMENT_USER_PERMISSION_DETAIL, "PATCH"): validate_can_manage_experiment,
        (routes.EXPERIMENT_USER_PERMISSION_DETAIL, "DELETE"): validate_can_manage_experiment,
        (routes.USER_REGISTERED_MODEL_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_USER_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_USER_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_USER_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_USER_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_USER_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_USER_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.PROMPT_USER_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.PROMPT_USER_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.PROMPT_USER_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.PROMPT_USER_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.PROMPT_USER_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.PROMPT_USER_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.USER_EXPERIMENT_PATTERN_PERMISSIONS, "GET"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PATTERN_PERMISSIONS, "POST"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PATTERN_PERMISSION_DETAIL, "GET"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PATTERN_PERMISSION_DETAIL, "POST"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PATTERN_PERMISSION_DETAIL, "PATCH"): validate_can_manage_experiment,
        (routes.USER_EXPERIMENT_PATTERN_PERMISSION_DETAIL, "DELETE"): validate_can_manage_experiment,
        (routes.USER_REGISTERED_MODEL_PATTERN_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PATTERN_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.USER_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PATTERN_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PATTERN_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PATTERN_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PATTERN_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PATTERN_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.USER_PROMPT_PATTERN_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.GROUP_EXPERIMENT_PERMISSIONS, "GET"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PERMISSIONS, "POST"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PERMISSION_DETAIL, "GET"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PERMISSION_DETAIL, "POST"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PERMISSION_DETAIL, "PATCH"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PERMISSION_DETAIL, "DELETE"): validate_can_manage_experiment,
        (routes.EXPERIMENT_GROUP_PERMISSIONS, "GET"): validate_can_manage_experiment,
        (routes.EXPERIMENT_GROUP_PERMISSIONS, "POST"): validate_can_manage_experiment,
        (routes.EXPERIMENT_GROUP_PERMISSION_DETAIL, "GET"): validate_can_manage_experiment,
        (routes.EXPERIMENT_GROUP_PERMISSION_DETAIL, "POST"): validate_can_manage_experiment,
        (routes.EXPERIMENT_GROUP_PERMISSION_DETAIL, "PATCH"): validate_can_manage_experiment,
        (routes.EXPERIMENT_GROUP_PERMISSION_DETAIL, "DELETE"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PATTERN_PERMISSIONS, "GET"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PATTERN_PERMISSIONS, "POST"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PATTERN_PERMISSION_DETAIL, "GET"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PATTERN_PERMISSION_DETAIL, "POST"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PATTERN_PERMISSION_DETAIL, "PATCH"): validate_can_manage_experiment,
        (routes.GROUP_EXPERIMENT_PATTERN_PERMISSION_DETAIL, "DELETE"): validate_can_manage_experiment,
        (routes.GROUP_REGISTERED_MODEL_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_GROUP_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_GROUP_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_GROUP_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_GROUP_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_GROUP_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.REGISTERED_MODEL_GROUP_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.PROMPT_GROUP_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.PROMPT_GROUP_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.PROMPT_GROUP_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.PROMPT_GROUP_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.PROMPT_GROUP_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.PROMPT_GROUP_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PATTERN_PERMISSIONS, "GET"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PATTERN_PERMISSIONS, "POST"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PATTERN_PERMISSION_DETAIL, "GET"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PATTERN_PERMISSION_DETAIL, "POST"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PATTERN_PERMISSION_DETAIL, "PATCH"): validate_can_manage_registered_model,
        (routes.GROUP_PROMPT_PATTERN_PERMISSION_DETAIL, "DELETE"): validate_can_manage_registered_model,
    }
)


LOGGED_MODEL_BEFORE_REQUEST_HANDLERS = {
    CreateLoggedModel: validate_can_update_experiment,
    GetLoggedModel: validate_can_read_logged_model,
    DeleteLoggedModel: validate_can_delete_logged_model,
    FinalizeLoggedModel: validate_can_update_logged_model,
    DeleteLoggedModelTag: validate_can_delete_logged_model,
    SetLoggedModelTags: validate_can_update_logged_model,
    LogLoggedModelParamsRequest: validate_can_update_logged_model,
}


def get_logged_model_before_request_handler(request_class):
    return LOGGED_MODEL_BEFORE_REQUEST_HANDLERS.get(request_class)


def _re_compile_path(path: str) -> re.Pattern:
    """
    Convert a path with angle brackets to a regex pattern. For example,
    "/api/2.0/experiments/<experiment_id>" becomes "/api/2.0/experiments/([^/]+)".
    """
    return re.compile(re.sub(r"<([^>]+)>", r"([^/]+)", path))


LOGGED_MODEL_BEFORE_REQUEST_VALIDATORS = {
    # Paths for logged models contains path parameters (e.g. /mlflow/logged-models/<model_id>)
    (_re_compile_path(http_path), method): handler
    for http_path, handler, methods in get_endpoints(get_logged_model_before_request_handler)
    for method in methods
}


def _get_proxy_artifact_validator(method: str, view_args: Optional[Dict[str, Any]]) -> Optional[Callable[[], bool]]:
    if view_args is None:
        return validate_can_read_experiment_artifact_proxy  # List

    return {
        "GET": validate_can_read_experiment_artifact_proxy,  # Download
        "PUT": validate_can_update_experiment_artifact_proxy,  # Upload
        "DELETE": validate_can_delete_experiment_artifact_proxy,  # Delete
    }.get(method)


def _is_proxy_artifact_path(path: str) -> bool:
    return path.startswith(f"{_REST_API_PATH_PREFIX}/mlflow-artifacts/artifacts/")


def _find_validator(req: Request) -> Optional[Callable[[], bool]]:
    """
    Finds the validator matching the request path and method.
    """
    if "/mlflow/logged-models" in req.path:
        # logged model routes are not registered in the app
        # so we need to check them manually
        return next(
            (v for (pat, method), v in LOGGED_MODEL_BEFORE_REQUEST_VALIDATORS.items() if pat.fullmatch(req.path) and method == req.method),
            None,
        )
    else:
        return BEFORE_REQUEST_VALIDATORS.get((req.path, req.method))


def before_request_hook():
    """Called before each request. If it did not return a response,
    the view function for the matched route is called and returns a response"""
    if _is_unprotected_route(request.path):
        return
    if request.authorization is not None:
        if request.authorization.type == "basic":
            if not authenticate_request_basic_auth():
                return responses.make_basic_auth_response()
        if request.authorization.type == "bearer":
            if not authenticate_request_bearer_token():
                return responses.make_auth_required_response()
    else:
        if session.get("username") is None:
            session.clear()

            if config.AUTOMATIC_LOGIN_REDIRECT:
                return redirect(url_for("login"))
            return render_template(
                "auth.html",
                username=None,
                provide_display_name=config.OIDC_PROVIDER_DISPLAY_NAME,
            )
    # admins don't need to be authorized
    if get_is_admin():
        return
    # authorization
    if validator := _find_validator(request):
        if not validator():
            return responses.make_forbidden_response()
    elif _is_proxy_artifact_path(request.path):
        if validator := _get_proxy_artifact_validator(request.method, request.view_args):
            if not validator():
                return responses.make_forbidden_response()



from flask import jsonify
from mlflow.server.handlers import catch_mlflow_exception

from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.utils import check_admin_permission, get_request_param


@catch_mlflow_exception
@check_admin_permission
def create_prompt_regex_permission(username: str):
    store.create_prompt_regex_permission(
        regex=get_request_param("regex"),
        priority=int(get_request_param("priority")),
        permission=get_request_param("permission"),
        username=username,
    )
    return jsonify({"status": "success"}), 200


@catch_mlflow_exception
@check_admin_permission
def list_prompt_regex_permissions(username: str):
    rm = store.list_prompt_regex_permissions(
        username=username,
    )
    return jsonify([r.to_json() for r in rm]), 200


@catch_mlflow_exception
@check_admin_permission
def get_prompt_regex_permission(username: str, pattern_id: str):
    rm = store.get_prompt_regex_permission(
        id=int(pattern_id),
        username=username,
    )
    return jsonify({"prompt_permission": rm.to_json()}), 200


@catch_mlflow_exception
@check_admin_permission
def update_prompt_regex_permission(username: str, pattern_id: str):
    rm = store.update_prompt_regex_permission(
        id=int(pattern_id),
        regex=get_request_param("regex"),
        priority=int(get_request_param("priority")),
        permission=get_request_param("permission"),
        username=username,
    )
    return jsonify({"prompt_permission": rm.to_json()}), 200


@catch_mlflow_exception
@check_admin_permission
def delete_prompt_regex_permission(username: str, pattern_id: str):
    store.delete_prompt_regex_permission(
        id=int(pattern_id),
        username=username,
    )
    return jsonify({"status": "success"}), 200



from mlflow.server.handlers import _get_rest_path

HOME = "/"
LOGIN = "/login"
LOGOUT = "/logout"
CALLBACK = "/callback"

STATIC = "/oidc/static/<path:filename>"
UI = "/oidc/ui/<path:filename>"
UI_ROOT = "/oidc/ui/"

# Runtime configuration endpoint under UI path
UI_CONFIG = "/oidc/ui/config.json"

########### API refactoring ###########
# USER, EXPERIMENT, PATTERN
USER_EXPERIMENT_PERMISSIONS = _get_rest_path("/mlflow/permissions/users/<string:username>/experiments")
USER_EXPERIMENT_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/users/<string:username>/experiments/<string:experiment_id>")

EXPERIMENT_USER_PERMISSIONS = _get_rest_path("/mlflow/permissions/experiments/<string:experiment_id>/users")
EXPERIMENT_USER_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/experiments/<string:experiment_id>/users/<string:username>")

USER_EXPERIMENT_PATTERN_PERMISSIONS = _get_rest_path("/mlflow/permissions/users/<string:username>/experiment-patterns")
USER_EXPERIMENT_PATTERN_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/users/<string:username>/experiment-patterns/<string:pattern_id>")

# USER, REGISTERED_MODEL, PATTERN
USER_REGISTERED_MODEL_PERMISSIONS = _get_rest_path("/mlflow/permissions/users/<string:username>/registered-models")
USER_REGISTERED_MODEL_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/users/<string:username>/registered-models/<string:name>")

REGISTERED_MODEL_USER_PERMISSIONS = _get_rest_path("/mlflow/permissions/registered-models/<string:name>/users")
REGISTERED_MODEL_USER_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/registered-models/<string:name>/users/<string:username>")

USER_REGISTERED_MODEL_PATTERN_PERMISSIONS = _get_rest_path("/mlflow/permissions/users/<string:username>/registered-models-patterns")
USER_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/users/<string:username>/registered-models-patterns/<string:pattern_id>")

# USER, PROMPT, PATTERN
USER_PROMPT_PERMISSIONS = _get_rest_path("/mlflow/permissions/users/<string:username>/prompts")
USER_PROMPT_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/users/<string:username>/prompts/<string:prompt_name>")

PROMPT_USER_PERMISSIONS = _get_rest_path("/mlflow/permissions/prompts/<string:prompt_name>/users")
PROMPT_USER_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/prompts/<string:prompt_name>/users/<string:username>")

USER_PROMPT_PATTERN_PERMISSIONS = _get_rest_path("/mlflow/permissions/users/<string:username>/prompts-patterns")
USER_PROMPT_PATTERN_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/users/<string:username>/prompts-patterns/<string:pattern_id>")

# GROUP STUFF

# GROUP -> EXPERIMENT, REGISTERED_MODEL, PROMPT
# GROUP, EXPERIMENT, PATTERN
GROUP_EXPERIMENT_PERMISSIONS = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/experiments")
GROUP_EXPERIMENT_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/experiments/<string:experiment_id>")

EXPERIMENT_GROUP_PERMISSIONS = _get_rest_path("/mlflow/permissions/experiments/<string:experiment_id>/groups")
EXPERIMENT_GROUP_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/experiments/<string:experiment_id>/groups/<string:group_name>")

GROUP_EXPERIMENT_PATTERN_PERMISSIONS = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/experiment-patterns")
GROUP_EXPERIMENT_PATTERN_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/experiment-patterns/<string:pattern_id>")

# GROUP, REGISTERED_MODEL, PATTERN
GROUP_REGISTERED_MODEL_PERMISSIONS = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/registered-models")
GROUP_REGISTERED_MODEL_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/registered-models/<string:name>")

REGISTERED_MODEL_GROUP_PERMISSIONS = _get_rest_path("/mlflow/permissions/registered-models/<string:name>/groups")
REGISTERED_MODEL_GROUP_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/registered-models/<string:name>/groups/<string:group_name>")

GROUP_REGISTERED_MODEL_PATTERN_PERMISSIONS = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/registered-models-patterns")
GROUP_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL = _get_rest_path(
    "/mlflow/permissions/groups/<string:group_name>/registered-models-patterns/<string:pattern_id>"
)

# GROUP, PROMPT, PATTERN
GROUP_PROMPT_PERMISSIONS = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/prompts")
GROUP_PROMPT_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/prompts/<string:prompt_name>")

PROMPT_GROUP_PERMISSIONS = _get_rest_path("/mlflow/permissions/prompts/<string:prompt_name>/groups")
PROMPT_GROUP_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/prompts/<string:prompt_name>/groups/<string:group_name>")

GROUP_PROMPT_PATTERN_PERMISSIONS = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/prompts-patterns")
GROUP_PROMPT_PATTERN_PERMISSION_DETAIL = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/prompts-patterns/<string:pattern_id>")

#######################################

# List of Resources

LIST_EXPERIMENTS = _get_rest_path("/mlflow/permissions/experiments")
LIST_PROMPTS = _get_rest_path("/mlflow/permissions/prompts")
LIST_MODELS = _get_rest_path("/mlflow/permissions/registered-models")
LIST_USERS = _get_rest_path("/mlflow/permissions/users")
LIST_GROUPS = _get_rest_path("/mlflow/permissions/groups")

GROUP_USER_PERMISSIONS = _get_rest_path("/mlflow/permissions/groups/<string:group_name>/users")
###############


# create access token for current user
CREATE_ACCESS_TOKEN = _get_rest_path("/mlflow/permissions/users/access-token")
# get infrmation about current user
GET_CURRENT_USER = _get_rest_path("/mlflow/permissions/users/current")

# CRUD routes from basic_auth
CREATE_USER = _get_rest_path("/mlflow/users/create")
GET_USER = _get_rest_path("/mlflow/users/get")
UPDATE_USER_PASSWORD = _get_rest_path("/mlflow/users/update-password")
UPDATE_USER_ADMIN = _get_rest_path("/mlflow/users/update-admin")
DELETE_USER = _get_rest_path("/mlflow/users/delete")



from mlflow_oidc_auth.permissions import Permission
from mlflow_oidc_auth.utils import effective_registered_model_permission, effective_experiment_permission, get_username, get_model_name, get_model_id
from mlflow.server.handlers import _get_tracking_store


def _get_permission_from_registered_model_name() -> Permission:
    model_name = get_model_name()
    username = get_username()
    return effective_registered_model_permission(model_name, username).permission


def _get_permission_from_model_id() -> Permission:
    # logged model permissions inherit from parent resource (experiment)
    model_id = get_model_id()
    model = _get_tracking_store().get_logged_model(model_id)
    experiment_id = model.experiment_id
    username = get_username()
    return effective_experiment_permission(experiment_id, username).permission


def validate_can_read_registered_model():
    return _get_permission_from_registered_model_name().can_read


def validate_can_update_registered_model():
    return _get_permission_from_registered_model_name().can_update


def validate_can_delete_registered_model():
    return _get_permission_from_registered_model_name().can_delete


def validate_can_manage_registered_model():
    return _get_permission_from_registered_model_name().can_manage


def validate_can_read_logged_model():
    return _get_permission_from_model_id().can_read


def validate_can_update_logged_model():
    return _get_permission_from_model_id().can_update


def validate_can_delete_logged_model():
    return _get_permission_from_model_id().can_delete


def validate_can_manage_logged_model():
    return _get_permission_from_model_id().can_manage



"""
Tests for dynamic OIDC redirect URI calculation.

These tests verify that the redirect URI is correctly calculated
from request headers in various proxy scenarios using ProxyFix middleware.
"""

import unittest
from unittest.mock import Mock, patch
from flask import Flask
from mlflow_oidc_auth.utils.uri_helpers import _get_dynamic_redirect_uri, get_configured_or_dynamic_redirect_uri, _get_base_url_from_request, normalize_url_port


class TestDynamicRedirectUri(unittest.TestCase):
    """Test cases for dynamic redirect URI calculation."""

    def setUp(self):
        """Set up test fixtures."""
        from werkzeug.middleware.proxy_fix import ProxyFix

        self.app = Flask(__name__)
        self.app.wsgi_app = ProxyFix(self.app.wsgi_app, x_proto=1, x_host=1, x_prefix=1)
        self.app.config["TESTING"] = True
        self.app_context = self.app.app_context()
        self.app_context.push()

    def tearDown(self):
        """Clean up test fixtures."""
        self.app_context.pop()

    def test_script_root_base(self):
        """Test redirect URI calculation using only request.script_root for base path."""
        cases = [
            # (base_url, script_root, expected_redirect)
            ("http://localhost:5000/", "", "http://localhost:5000/callback"),
            ("https://example.com/mlflow/", "/mlflow", "https://example.com/callback"),
            ("https://corp.example.com/apps/ml-platform/", "/apps/ml-platform", "https://corp.example.com/callback"),
            ("http://localhost:5000/myapp/", "/myapp", "http://localhost:5000/callback"),
            ("https://k8s-cluster.example.com/v1/ml-platform/", "/v1/ml-platform", "https://k8s-cluster.example.com/callback"),
            ("https://example.com:8443/my-app/", "/my-app", "https://example.com:8443/callback"),
            ("http://localhost:8080/", "", "http://localhost:8080/callback"),
            ("http://localhost:5000/", "", "http://localhost:5000/callback"),
            ("https://example.com:443/my-app/", "/my-app", "https://example.com/callback"),
            ("http://example.com:80/my-app/", "/my-app", "http://example.com/callback"),
        ]
        for url, script_root, expected in cases:
            # Ensure the URL path matches SCRIPT_NAME for correct script_root
            if script_root:
                # Ensure the path in the URL starts with script_root
                url = url.rstrip("/") + script_root + "/"
            environ_base = {"SCRIPT_NAME": script_root} if script_root else {}
            with self.app.test_request_context(url, environ_base=environ_base):
                result = get_configured_or_dynamic_redirect_uri(None)
                self.assertEqual(result, expected)

    def test_http_standard_port_omitted(self):
        """Test redirect URI calculation where HTTP standard port 80 is omitted."""
        url = "http://example.com:80/my-app/"
        with self.app.test_request_context(url, environ_base={"SCRIPT_NAME": "/my-app"}):
            result = _get_dynamic_redirect_uri()
            expected = "http://example.com/callback"
            self.assertEqual(result, expected)

    def test_get_base_url_from_request(self):
        """Test getting base URL from request context."""
        url = "https://example.com/my-app/"
        with self.app.test_request_context(url, environ_base={"SCRIPT_NAME": "/my-app"}):
            result = _get_base_url_from_request()
            expected = "https://example.com"
            self.assertEqual(result, expected)


class TestPortNormalization(unittest.TestCase):
    """Test cases for URL port normalization."""

    def test_normalize_https_standard_port(self):
        """Test that HTTPS port 443 is omitted."""
        url = "https://example.com:443/path"
        result = normalize_url_port(url)
        expected = "https://example.com/path"
        self.assertEqual(result, expected)

    def test_normalize_http_standard_port(self):
        """Test that HTTP port 80 is omitted."""
        url = "http://example.com:80/path"
        result = normalize_url_port(url)
        expected = "http://example.com/path"
        self.assertEqual(result, expected)

    def test_preserve_custom_https_port(self):
        """Test that custom HTTPS ports are preserved."""
        url = "https://example.com:8443/path"
        result = normalize_url_port(url)
        expected = "https://example.com:8443/path"
        self.assertEqual(result, expected)

    def test_preserve_custom_http_port(self):
        """Test that custom HTTP ports are preserved."""
        url = "http://example.com:8080/path"
        result = normalize_url_port(url)
        expected = "http://example.com:8080/path"
        self.assertEqual(result, expected)

    def test_no_port_in_url(self):
        """Test that URLs without explicit ports are unchanged."""
        url = "https://example.com/path"
        result = normalize_url_port(url)
        expected = "https://example.com/path"
        self.assertEqual(result, expected)

    def test_localhost_custom_port(self):
        """Test that localhost custom ports are preserved."""
        url = "http://localhost:5000/path"
        result = normalize_url_port(url)
        expected = "http://localhost:5000/path"
        self.assertEqual(result, expected)

    def test_malformed_url_handling(self):
        """Test that malformed URLs are returned unchanged."""
        url = "not-a-valid-url"
        result = normalize_url_port(url)
        expected = "not-a-valid-url"
        self.assertEqual(result, expected)

    def test_url_with_userinfo_and_standard_port(self):
        """Test that URLs with userinfo and standard ports are handled correctly."""
        url = "https://user:pass@example.com:443/path"
        result = normalize_url_port(url)
        expected = "https://user:pass@example.com/path"
        self.assertEqual(result, expected)

    def test_url_with_userinfo_and_custom_port(self):
        """Test that URLs with userinfo and custom ports preserve the port."""
        url = "https://user:pass@example.com:8443/path"
        result = normalize_url_port(url)
        expected = "https://user:pass@example.com:8443/path"
        self.assertEqual(result, expected)


class TestRequestContextRequirement(unittest.TestCase):
    """Test cases for functions that require Flask request context."""

    def setUp(self):
        """Set up test fixtures."""
        self.app = Flask(__name__)
        self.app.config["TESTING"] = True

    def test_get_base_url_from_request_no_context(self):
        """Test that _get_base_url_from_request raises RuntimeError without request context."""
        with self.assertRaises(RuntimeError) as context:
            _get_base_url_from_request()
        self.assertIn("requires an active Flask request context", str(context.exception))

    def test_get_dynamic_redirect_uri_no_context(self):
        """Test that _get_dynamic_redirect_uri raises RuntimeError without request context."""
        with self.assertRaises(RuntimeError) as context:
            _get_dynamic_redirect_uri()
        self.assertIn("requires an active Flask request context", str(context.exception))

    def test_get_dynamic_redirect_uri_empty_callback_path(self):
        """Test redirect URI calculation with empty callback path."""
        with self.app.test_request_context("http://localhost:5000/"):
            result = _get_dynamic_redirect_uri("")
            expected = "http://localhost:5000/"
            self.assertEqual(result, expected)

    def test_configured_or_dynamic_redirect_uri_whitespace_config(self):
        """Test that whitespace-only configured URI falls back to dynamic calculation."""
        with self.app.test_request_context("http://localhost:5000/"):
            result = get_configured_or_dynamic_redirect_uri("   ")
            expected = "http://localhost:5000/callback"
            self.assertEqual(result, expected)

    def test_configured_or_dynamic_redirect_uri_empty_string_config(self):
        """Test that empty string configured URI falls back to dynamic calculation."""
        with self.app.test_request_context("http://localhost:5000/"):
            result = get_configured_or_dynamic_redirect_uri("")
            expected = "http://localhost:5000/callback"
            self.assertEqual(result, expected)


if __name__ == "__main__":
    unittest.main()



from flask import jsonify
from mlflow.server.handlers import catch_mlflow_exception

from mlflow_oidc_auth.store import store


@catch_mlflow_exception
def list_groups():
    return store.get_groups()


@catch_mlflow_exception
def get_group_users(group_name):
    return jsonify({"users": store.get_group_users(group_name)})



import os

from flask_caching import Cache
from flask_session import Session
from mlflow.server import app
from werkzeug.middleware.proxy_fix import ProxyFix

from mlflow_oidc_auth import routes, views
from mlflow_oidc_auth.config import config
from mlflow_oidc_auth.hooks import after_request_hook, before_request_hook
from mlflow_oidc_auth.logger import get_logger

logger = get_logger()

# Configure custom Flask app
template_dir = os.path.dirname(__file__)
template_dir = os.path.join(template_dir, "templates")

app.config.from_object(config)
app.secret_key = app.config["SECRET_KEY"].encode("utf8")
app.template_folder = template_dir
static_folder = app.static_folder

# Configure ProxyFix middleware to handle reverse proxy headers
app.wsgi_app = ProxyFix(
    app.wsgi_app,
    x_for=config.PROXY_FIX_X_FOR,
    x_proto=config.PROXY_FIX_X_PROTO,
    x_host=config.PROXY_FIX_X_HOST,
    x_port=config.PROXY_FIX_X_PORT,
    x_prefix=config.PROXY_FIX_X_PREFIX,
)

logger.debug(
    f"ProxyFix middleware configured - x_for={config.PROXY_FIX_X_FOR}, x_proto={config.PROXY_FIX_X_PROTO}, x_host={config.PROXY_FIX_X_HOST}, x_port={config.PROXY_FIX_X_PORT}, x_prefix={config.PROXY_FIX_X_PREFIX}"
)

# Add links to MLFlow UI
if config.EXTEND_MLFLOW_MENU:
    app.view_functions["serve"] = views.index

# OIDC routes
app.add_url_rule(rule=routes.LOGIN, methods=["GET"], view_func=views.login)
app.add_url_rule(rule=routes.LOGOUT, methods=["GET"], view_func=views.logout)
app.add_url_rule(rule=routes.CALLBACK, methods=["GET"], view_func=views.callback)

# UI routes
app.add_url_rule(rule=routes.STATIC, methods=["GET"], view_func=views.oidc_static)
app.add_url_rule(rule=routes.UI, methods=["GET"], view_func=views.oidc_ui)
app.add_url_rule(rule=routes.UI_ROOT, methods=["GET"], view_func=views.oidc_ui)

# Runtime configuration endpoint under UI path
app.add_url_rule(rule=routes.UI_CONFIG, methods=["GET"], view_func=views.get_runtime_config)

# User token
app.add_url_rule(rule=routes.CREATE_ACCESS_TOKEN, methods=["PATCH"], view_func=views.create_user_access_token)
app.add_url_rule(rule=routes.GET_CURRENT_USER, methods=["GET"], view_func=views.get_current_user)

# User management
app.add_url_rule(rule=routes.CREATE_USER, methods=["POST"], view_func=views.create_new_user)
app.add_url_rule(rule=routes.GET_USER, methods=["GET"], view_func=views.get_user)
app.add_url_rule(rule=routes.UPDATE_USER_PASSWORD, methods=["PATCH"], view_func=views.update_username_password)
app.add_url_rule(rule=routes.UPDATE_USER_ADMIN, methods=["PATCH"], view_func=views.update_user_admin)
app.add_url_rule(rule=routes.DELETE_USER, methods=["DELETE"], view_func=views.delete_user)


# UI routes support
app.add_url_rule(rule=routes.EXPERIMENT_USER_PERMISSIONS, methods=["GET"], view_func=views.get_experiment_users)
app.add_url_rule(rule=routes.PROMPT_USER_PERMISSIONS, methods=["GET"], view_func=views.get_prompt_users)
app.add_url_rule(rule=routes.REGISTERED_MODEL_USER_PERMISSIONS, methods=["GET"], view_func=views.get_registered_model_users)

# List resources
app.add_url_rule(rule=routes.LIST_EXPERIMENTS, methods=["GET"], view_func=views.list_experiments)
app.add_url_rule(rule=routes.LIST_MODELS, methods=["GET"], view_func=views.list_registered_models)
app.add_url_rule(rule=routes.LIST_PROMPTS, methods=["GET"], view_func=views.list_prompts)
app.add_url_rule(rule=routes.LIST_USERS, methods=["GET"], view_func=views.list_users)
app.add_url_rule(rule=routes.LIST_GROUPS, methods=["GET"], view_func=views.list_groups)

# user experiment permission management
app.add_url_rule(rule=routes.USER_EXPERIMENT_PERMISSIONS, methods=["GET"], view_func=views.list_user_experiments)
app.add_url_rule(rule=routes.USER_EXPERIMENT_PERMISSION_DETAIL, methods=["POST"], view_func=views.create_experiment_permission)
app.add_url_rule(rule=routes.USER_EXPERIMENT_PERMISSION_DETAIL, methods=["GET"], view_func=views.get_experiment_permission)
app.add_url_rule(rule=routes.USER_EXPERIMENT_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_experiment_permission)
app.add_url_rule(rule=routes.USER_EXPERIMENT_PERMISSION_DETAIL, methods=["DELETE"], view_func=views.delete_experiment_permission)

# user experiment regex permission management
app.add_url_rule(rule=routes.USER_EXPERIMENT_PATTERN_PERMISSIONS, methods=["POST"], view_func=views.create_experiment_regex_permission)
app.add_url_rule(rule=routes.USER_EXPERIMENT_PATTERN_PERMISSIONS, methods=["GET"], view_func=views.list_user_experiment_regex_permission)
app.add_url_rule(rule=routes.USER_EXPERIMENT_PATTERN_PERMISSION_DETAIL, methods=["GET"], view_func=views.get_experiment_regex_permission)
app.add_url_rule(rule=routes.USER_EXPERIMENT_PATTERN_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_experiment_regex_permission)
app.add_url_rule(
    rule=routes.USER_EXPERIMENT_PATTERN_PERMISSION_DETAIL,
    methods=["DELETE"],
    view_func=views.delete_experiment_regex_permission,
)

# user prompt management
app.add_url_rule(rule=routes.USER_PROMPT_PERMISSIONS, methods=["GET"], view_func=views.list_user_prompts)
app.add_url_rule(rule=routes.USER_PROMPT_PERMISSION_DETAIL, methods=["POST"], view_func=views.create_prompt_permission)
app.add_url_rule(rule=routes.USER_PROMPT_PERMISSION_DETAIL, methods=["GET"], view_func=views.get_prompt_permission)
app.add_url_rule(rule=routes.USER_PROMPT_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_prompt_permission)
app.add_url_rule(rule=routes.USER_PROMPT_PERMISSION_DETAIL, methods=["DELETE"], view_func=views.delete_prompt_permission)

# user prompt regex permission management
app.add_url_rule(rule=routes.USER_PROMPT_PATTERN_PERMISSIONS, methods=["GET"], view_func=views.list_prompt_regex_permissions)
app.add_url_rule(rule=routes.USER_PROMPT_PATTERN_PERMISSIONS, methods=["POST"], view_func=views.create_prompt_regex_permission)
app.add_url_rule(rule=routes.USER_PROMPT_PATTERN_PERMISSION_DETAIL, methods=["GET"], view_func=views.get_prompt_regex_permission)
app.add_url_rule(rule=routes.USER_PROMPT_PATTERN_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_prompt_regex_permission)
app.add_url_rule(rule=routes.USER_PROMPT_PATTERN_PERMISSION_DETAIL, methods=["DELETE"], view_func=views.delete_prompt_regex_permission)

# user registered model management
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PERMISSIONS, methods=["GET"], view_func=views.list_user_models)
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PERMISSION_DETAIL, methods=["POST"], view_func=views.create_registered_model_permission)
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PERMISSION_DETAIL, methods=["GET"], view_func=views.get_registered_model_permission)
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_registered_model_permission)
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PERMISSION_DETAIL, methods=["DELETE"], view_func=views.delete_registered_model_permission)

# user registered model regex permission management
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PATTERN_PERMISSIONS, methods=["GET"], view_func=views.list_registered_model_regex_permissions)
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PATTERN_PERMISSIONS, methods=["POST"], view_func=views.create_registered_model_regex_permission)
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, methods=["GET"], view_func=views.get_registered_model_regex_permission)
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_registered_model_regex_permission)
app.add_url_rule(rule=routes.USER_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL, methods=["DELETE"], view_func=views.delete_registered_model_regex_permission)

app.add_url_rule(rule=routes.GROUP_USER_PERMISSIONS, methods=["GET"], view_func=views.get_group_users)

app.add_url_rule(rule=routes.GROUP_EXPERIMENT_PERMISSIONS, methods=["GET"], view_func=views.list_group_experiments)
app.add_url_rule(rule=routes.GROUP_EXPERIMENT_PERMISSION_DETAIL, methods=["POST"], view_func=views.create_group_experiment_permission)
app.add_url_rule(rule=routes.GROUP_EXPERIMENT_PERMISSION_DETAIL, methods=["DELETE"], view_func=views.delete_group_experiment_permission)
app.add_url_rule(rule=routes.GROUP_EXPERIMENT_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_group_experiment_permission)

app.add_url_rule(rule=routes.GROUP_REGISTERED_MODEL_PERMISSIONS, methods=["GET"], view_func=views.list_group_models)
app.add_url_rule(rule=routes.GROUP_REGISTERED_MODEL_PERMISSION_DETAIL, methods=["POST"], view_func=views.create_group_model_permission)
app.add_url_rule(rule=routes.GROUP_REGISTERED_MODEL_PERMISSION_DETAIL, methods=["DELETE"], view_func=views.delete_group_model_permission)
app.add_url_rule(rule=routes.GROUP_REGISTERED_MODEL_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_group_model_permission)

app.add_url_rule(rule=routes.GROUP_PROMPT_PERMISSIONS, methods=["GET"], view_func=views.get_group_prompts)
app.add_url_rule(rule=routes.GROUP_PROMPT_PERMISSION_DETAIL, methods=["POST"], view_func=views.create_group_prompt_permission)
app.add_url_rule(rule=routes.GROUP_PROMPT_PERMISSION_DETAIL, methods=["DELETE"], view_func=views.delete_group_prompt_permission)
app.add_url_rule(rule=routes.GROUP_PROMPT_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_group_prompt_permission)

app.add_url_rule(rule=routes.GROUP_EXPERIMENT_PATTERN_PERMISSIONS, methods=["GET"], view_func=views.list_group_experiment_regex_permissions)
app.add_url_rule(rule=routes.GROUP_EXPERIMENT_PATTERN_PERMISSIONS, methods=["POST"], view_func=views.create_group_experiment_regex_permission)
app.add_url_rule(rule=routes.GROUP_EXPERIMENT_PATTERN_PERMISSION_DETAIL, methods=["GET"], view_func=views.get_group_experiment_regex_permission)
app.add_url_rule(
    rule=routes.GROUP_EXPERIMENT_PATTERN_PERMISSION_DETAIL,
    methods=["PATCH"],
    view_func=views.update_group_experiment_regex_permission,
)
app.add_url_rule(
    rule=routes.GROUP_EXPERIMENT_PATTERN_PERMISSION_DETAIL,
    methods=["DELETE"],
    view_func=views.delete_group_experiment_regex_permission,
)

app.add_url_rule(
    rule=routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSIONS,
    methods=["POST"],
    view_func=views.create_group_registered_model_regex_permission,
)
app.add_url_rule(
    rule=routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSIONS,
    methods=["GET"],
    view_func=views.list_group_registered_model_regex_permissions,
)
app.add_url_rule(
    rule=routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL,
    methods=["GET"],
    view_func=views.get_group_registered_model_regex_permission,
)
app.add_url_rule(
    rule=routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL,
    methods=["PATCH"],
    view_func=views.update_group_registered_model_regex_permission,
)
app.add_url_rule(
    rule=routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL,
    methods=["DELETE"],
    view_func=views.delete_group_registered_model_regex_permission,
)

app.add_url_rule(rule=routes.GROUP_PROMPT_PATTERN_PERMISSIONS, methods=["GET"], view_func=views.list_group_prompt_regex_permissions)
app.add_url_rule(rule=routes.GROUP_PROMPT_PATTERN_PERMISSIONS, methods=["POST"], view_func=views.create_group_prompt_regex_permission)
app.add_url_rule(rule=routes.GROUP_PROMPT_PATTERN_PERMISSION_DETAIL, methods=["GET"], view_func=views.get_group_prompt_regex_permission)
app.add_url_rule(rule=routes.GROUP_PROMPT_PATTERN_PERMISSION_DETAIL, methods=["PATCH"], view_func=views.update_group_prompt_regex_permission)
app.add_url_rule(rule=routes.GROUP_PROMPT_PATTERN_PERMISSION_DETAIL, methods=["DELETE"], view_func=views.delete_group_prompt_regex_permission)
###############################


# Add new hooks
app.before_request(before_request_hook)
app.after_request(after_request_hook)

# Set up session
Session(app)
cache = Cache(app)



from flask import jsonify, make_response
from mlflow.server.handlers import _get_model_registry_store, catch_mlflow_exception

from mlflow_oidc_auth.responses.client_error import make_forbidden_response
from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.utils import (
    can_manage_registered_model,
    check_registered_model_permission,
    fetch_all_prompts,
    get_is_admin,
    get_request_param,
    get_username,
)


@catch_mlflow_exception
@check_registered_model_permission
def create_prompt_permission(username: str, prompt_name: str):
    store.create_registered_model_permission(
        name=prompt_name,
        username=username,
        permission=get_request_param("permission"),
    )
    return jsonify({"message": "Model permission has been created."})


@catch_mlflow_exception
@check_registered_model_permission
def get_prompt_permission(username: str, prompt_name: str):
    rmp = store.get_registered_model_permission(prompt_name, username)
    return make_response({"prompt_permission": rmp.to_json()})


@catch_mlflow_exception
@check_registered_model_permission
def update_prompt_permission(username: str, prompt_name: str):
    store.update_registered_model_permission(prompt_name, username, get_request_param("permission"))
    return make_response(jsonify({"message": "Model permission has been changed"}))


@catch_mlflow_exception
@check_registered_model_permission
def delete_prompt_permission(username: str, prompt_name: str):
    store.delete_registered_model_permission(prompt_name, username)
    return make_response(jsonify({"message": "Model permission has been deleted"}))


# TODO: refactor it, move filtering logic to the store
@catch_mlflow_exception
def list_prompts():
    if get_is_admin():
        prompts = fetch_all_prompts()
    else:
        current_user = store.get_user(get_username())
        prompts = []
        for model in fetch_all_prompts():
            if can_manage_registered_model(model.name, current_user.username):
                prompts.append(model)
    models = [
        {
            "name": model.name,
            "tags": model.tags,
            "description": model.description,
            "aliases": model.aliases,
        }
        for model in prompts
    ]
    return jsonify(models)


@catch_mlflow_exception
def get_prompt_users(prompt_name):
    if not get_is_admin():
        current_user = store.get_user(get_username())
        if not can_manage_registered_model(prompt_name, current_user.username):
            return make_forbidden_response()
    list_users = store.list_users(all=True)
    # Filter users who are associated with the given model
    users = []
    for user in list_users:
        # Check if the user is associated with the model
        user_models = {model.name: model.permission for model in user.registered_model_permissions} if user.registered_model_permissions else {}
        if prompt_name in user_models:
            users.append(
                {
                    "username": user.username,
                    "permission": user_models[prompt_name],
                    "kind": "user" if not user.is_service_account else "service-account",
                }
            )
    return jsonify(users)



#!/usr/bin/env python3
"""
Comprehensive tests for URL port normalization functionality.

This test module validates that the normalize_url_port function correctly
handles various URL formats, port combinations, and edge cases.
"""

import unittest
from mlflow_oidc_auth.utils.uri_helpers import normalize_url_port


class TestPortNormalization(unittest.TestCase):
    """Test cases for URL port normalization functionality."""

    def test_normalize_https_standard_port(self):
        """Test that HTTPS port 443 is omitted from URLs."""
        url = "https://example.com:443/path"
        result = normalize_url_port(url)
        expected = "https://example.com/path"
        self.assertEqual(result, expected)

    def test_normalize_http_standard_port(self):
        """Test that HTTP port 80 is omitted from URLs."""
        url = "http://example.com:80/path"
        result = normalize_url_port(url)
        expected = "http://example.com/path"
        self.assertEqual(result, expected)

    def test_preserve_custom_https_port(self):
        """Test that custom HTTPS ports are preserved in URLs."""
        url = "https://example.com:8443/path"
        result = normalize_url_port(url)
        expected = "https://example.com:8443/path"
        self.assertEqual(result, expected)

    def test_preserve_custom_http_port(self):
        """Test that custom HTTP ports are preserved in URLs."""
        url = "http://example.com:8080/path"
        result = normalize_url_port(url)
        expected = "http://example.com:8080/path"
        self.assertEqual(result, expected)

    def test_no_port_in_url(self):
        """Test that URLs without explicit ports remain unchanged."""
        url = "https://example.com/path"
        result = normalize_url_port(url)
        expected = "https://example.com/path"
        self.assertEqual(result, expected)

    def test_localhost_custom_port(self):
        """Test that localhost custom ports are preserved."""
        url = "http://localhost:5000/path"
        result = normalize_url_port(url)
        expected = "http://localhost:5000/path"
        self.assertEqual(result, expected)

    def test_https_standard_port_with_mlflow_callback(self):
        """Test HTTPS standard port normalization with MLflow callback path."""
        url = "https://example.com:443/apps/mlflow/oidc/callback"
        result = normalize_url_port(url)
        expected = "https://example.com/apps/mlflow/oidc/callback"
        self.assertEqual(result, expected)

    def test_http_standard_port_with_mlflow_callback(self):
        """Test HTTP standard port normalization with MLflow callback path."""
        url = "http://example.com:80/apps/mlflow/oidc/callback"
        result = normalize_url_port(url)
        expected = "http://example.com/apps/mlflow/oidc/callback"
        self.assertEqual(result, expected)

    def test_malformed_url_handling(self):
        """Test that malformed URLs are returned unchanged without errors."""
        url = "not-a-valid-url"
        result = normalize_url_port(url)
        expected = "not-a-valid-url"
        self.assertEqual(result, expected)

    def test_url_with_userinfo_and_standard_port(self):
        """Test that URLs with userinfo and standard ports are handled correctly."""
        url = "https://user:pass@example.com:443/path"
        result = normalize_url_port(url)
        expected = "https://user:pass@example.com/path"
        self.assertEqual(result, expected)

    def test_url_with_userinfo_and_custom_port(self):
        """Test that URLs with userinfo and custom ports preserve the port."""
        url = "https://user:pass@example.com:8443/path"
        result = normalize_url_port(url)
        expected = "https://user:pass@example.com:8443/path"
        self.assertEqual(result, expected)

    def test_url_with_query_parameters(self):
        """Test that URLs with query parameters are handled correctly."""
        url = "https://example.com:443/path?param=value"
        result = normalize_url_port(url)
        expected = "https://example.com/path?param=value"
        self.assertEqual(result, expected)

    def test_url_with_fragment(self):
        """Test that URLs with fragments are handled correctly."""
        url = "https://example.com:443/path#section"
        result = normalize_url_port(url)
        expected = "https://example.com/path#section"
        self.assertEqual(result, expected)

    def test_edge_case_empty_string(self):
        """Test handling of empty string input."""
        url = ""
        result = normalize_url_port(url)
        expected = ""
        self.assertEqual(result, expected)

    def test_edge_case_none_input(self):
        """Test handling of None input (should handle gracefully)."""
        with self.assertRaises((TypeError, AttributeError)):
            normalize_url_port(None)


if __name__ == "__main__":
    unittest.main()



from functools import wraps
from typing import Callable

from mlflow_oidc_auth.logger import get_logger
from mlflow_oidc_auth.responses.client_error import make_forbidden_response
from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.utils.permissions import can_manage_experiment, can_manage_registered_model
from mlflow_oidc_auth.utils.request_helpers import get_experiment_id, get_is_admin, get_model_name, get_username

logger = get_logger()


def check_experiment_permission(f) -> Callable:
    @wraps(f)
    def decorated_function(*args, **kwargs):
        current_user = store.get_user(get_username())
        if not get_is_admin():
            logger.debug(f"Not Admin. Checking permission for {current_user.username}")
            experiment_id = get_experiment_id()
            if not can_manage_experiment(experiment_id, current_user.username):
                logger.warning(f"Change permission denied for {current_user.username} on experiment {experiment_id}")
                return make_forbidden_response()
        logger.debug(f"Change permission granted for {current_user.username}")
        return f(*args, **kwargs)

    return decorated_function


def check_registered_model_permission(f) -> Callable:
    @wraps(f)
    def decorated_function(*args, **kwargs):
        current_user = store.get_user(get_username())
        if not get_is_admin():
            logger.debug(f"Not Admin. Checking permission for {current_user.username}")
            model_name = get_model_name()
            if not can_manage_registered_model(model_name, current_user.username):
                logger.warning(f"Change permission denied for {current_user.username} on model {model_name}")
                return make_forbidden_response()
        logger.debug(f"Permission granted for {current_user.username}")
        return f(*args, **kwargs)

    return decorated_function


def check_prompt_permission(f) -> Callable:
    @wraps(f)
    def decorated_function(*args, **kwargs):
        current_user = store.get_user(get_username())
        if not get_is_admin():
            logger.debug(f"Not Admin. Checking permission for {current_user.username}")
            prompt_name = get_model_name()
            if not can_manage_registered_model(prompt_name, current_user.username):
                logger.warning(f"Change permission denied for {current_user.username} on prompt {prompt_name}")
                return make_forbidden_response()
        logger.debug(f"Permission granted for {current_user.username}")
        return f(*args, **kwargs)

    return decorated_function


def check_admin_permission(f) -> Callable:
    @wraps(f)
    def decorated_function(*args, **kwargs):
        current_user = store.get_user(get_username())
        if not get_is_admin():
            logger.warning(f"Admin permission denied for {current_user.username}")
            return make_forbidden_response()
        logger.debug(f"Admin permission granted for {current_user.username}")
        return f(*args, **kwargs)

    return decorated_function



import pytest
from unittest.mock import patch, MagicMock
from flask import Flask, session, request, Response
from mlflow_oidc_auth.hooks.before_request import before_request_hook
from mlflow_oidc_auth import responses
from mlflow_oidc_auth.config import config

app = Flask(__name__)
app.secret_key = "test_secret_key"


@pytest.fixture
def client():
    with app.test_client() as client:
        yield client


def test_unprotected_route(client):
    with app.test_request_context(path="/health", method="GET"):
        assert before_request_hook() is None  # No response for unprotected routes


def test_basic_auth_failure(client):
    with app.test_request_context(path="/protected", method="GET"):
        # Mock the request.authorization object
        mock_auth = MagicMock()
        mock_auth.type = "basic"

        with patch("mlflow_oidc_auth.hooks.before_request.request") as mock_request, patch(
            "mlflow_oidc_auth.hooks.before_request.authenticate_request_basic_auth", return_value=False
        ), patch("mlflow_oidc_auth.hooks.before_request.responses.make_basic_auth_response", return_value=Response("Unauthorized", status=401)):
            mock_request.path = "/protected"
            mock_request.method = "GET"
            mock_request.authorization = mock_auth

            response = before_request_hook()
            assert response.status_code == 401  # type: ignore
            assert b"Unauthorized" in response.data  # type: ignore


def test_bearer_auth_failure(client):
    with app.test_request_context(path="/protected", method="GET"):
        # Mock the request.authorization object
        mock_auth = MagicMock()
        mock_auth.type = "bearer"

        with patch("mlflow_oidc_auth.hooks.before_request.request") as mock_request, patch(
            "mlflow_oidc_auth.hooks.before_request.authenticate_request_bearer_token", return_value=False
        ), patch("mlflow_oidc_auth.hooks.before_request.responses.make_auth_required_response", return_value=Response("Unauthorized", status=401)):
            mock_request.path = "/protected"
            mock_request.method = "GET"
            mock_request.authorization = mock_auth

            response = before_request_hook()
            assert response.status_code == 401  # type: ignore


def test_session_redirect(client):
    with app.test_request_context(path="/protected", method="GET"):
        session.clear()
        with patch("mlflow_oidc_auth.hooks.before_request.config.AUTOMATIC_LOGIN_REDIRECT", True), patch(
            "mlflow_oidc_auth.hooks.before_request.url_for", return_value="/login"
        ):
            response = before_request_hook()
            assert response.status_code == 302  # type: ignore
            assert response.location.endswith("/login")  # type: ignore


def test_authorization_failure(client):
    with app.test_request_context(path="/protected", method="GET"):
        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request.BEFORE_REQUEST_VALIDATORS", {("/protected", "GET"): lambda: False}
        ), patch("mlflow_oidc_auth.hooks.before_request.render_template", return_value=Response("Forbidden", status=403)):
            response = before_request_hook()
            assert response.status_code == 403  # type: ignore
            assert b"Forbidden" in response.data  # type: ignore


def test_basic_auth_success(client):
    """Test successful basic authentication"""
    with app.test_request_context(path="/protected", method="GET"):
        # Mock the request.authorization object
        mock_auth = MagicMock()
        mock_auth.type = "basic"

        with patch("mlflow_oidc_auth.hooks.before_request.request") as mock_request, patch(
            "mlflow_oidc_auth.hooks.before_request.authenticate_request_basic_auth", return_value=True
        ), patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request.BEFORE_REQUEST_VALIDATORS", {("/protected", "GET"): lambda: True}
        ):
            mock_request.path = "/protected"
            mock_request.method = "GET"
            mock_request.authorization = mock_auth

            response = before_request_hook()
            assert response is None  # No response means authentication succeeded


def test_bearer_auth_success(client):
    """Test successful bearer token authentication"""
    with app.test_request_context(path="/protected", method="GET", headers={"Authorization": "Bearer valid"}):
        with patch("mlflow_oidc_auth.hooks.before_request.authenticate_request_bearer_token", return_value=True), patch(
            "mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False
        ), patch("mlflow_oidc_auth.hooks.before_request.BEFORE_REQUEST_VALIDATORS", {("/protected", "GET"): lambda: True}):
            response = before_request_hook()
            assert response is None  # No response means authentication succeeded


def test_session_no_redirect(client):
    """Test session authentication without automatic redirect"""
    with app.test_request_context(path="/protected", method="GET"):
        session.clear()
        with patch("mlflow_oidc_auth.hooks.before_request.config.AUTOMATIC_LOGIN_REDIRECT", False), patch(
            "mlflow_oidc_auth.hooks.before_request.render_template", return_value=Response("Auth required", status=200)
        ) as mock_render:
            response = before_request_hook()
            assert response.status_code == 200  # type: ignore
            mock_render.assert_called_once_with(
                "auth.html",
                username=None,
                provide_display_name=config.OIDC_PROVIDER_DISPLAY_NAME,
            )


def test_admin_bypass(client):
    """Test that admin users bypass authorization"""
    with app.test_request_context(path="/protected", method="GET"):
        session["username"] = "admin"
        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=True):
            response = before_request_hook()
            assert response is None  # Admin should bypass authorization


def test_authorization_success(client):
    """Test successful authorization"""
    with app.test_request_context(path="/protected", method="GET"):
        session["username"] = "user"
        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request.BEFORE_REQUEST_VALIDATORS", {("/protected", "GET"): lambda: True}
        ):
            response = before_request_hook()
            assert response is None  # Authorization succeeded


def test_find_validator_logged_models(client):
    """Test _find_validator for logged model routes"""
    from mlflow_oidc_auth.hooks.before_request import _find_validator

    mock_request = MagicMock()
    mock_request.path = "/api/2.0/mlflow/logged-models/12345"
    mock_request.method = "GET"

    mock_pattern = MagicMock()
    mock_pattern.fullmatch.return_value = True
    mock_validator = lambda: True

    with patch("mlflow_oidc_auth.hooks.before_request.LOGGED_MODEL_BEFORE_REQUEST_VALIDATORS", {(mock_pattern, "GET"): mock_validator}):
        result = _find_validator(mock_request)
        assert result == mock_validator


def test_find_validator_regular_routes(client):
    """Test _find_validator for regular routes"""
    from mlflow_oidc_auth.hooks.before_request import _find_validator

    mock_request = MagicMock()
    mock_request.path = "/api/2.0/mlflow/experiments/create"
    mock_request.method = "POST"

    mock_validator = lambda: True

    with patch("mlflow_oidc_auth.hooks.before_request.BEFORE_REQUEST_VALIDATORS", {("/api/2.0/mlflow/experiments/create", "POST"): mock_validator}):
        result = _find_validator(mock_request)
        assert result == mock_validator


def test_find_validator_no_match(client):
    """Test _find_validator when no validator is found"""
    from mlflow_oidc_auth.hooks.before_request import _find_validator

    mock_request = MagicMock()
    mock_request.path = "/unknown/path"
    mock_request.method = "GET"

    with patch("mlflow_oidc_auth.hooks.before_request.BEFORE_REQUEST_VALIDATORS", {}), patch(
        "mlflow_oidc_auth.hooks.before_request.LOGGED_MODEL_BEFORE_REQUEST_VALIDATORS", {}
    ):
        result = _find_validator(mock_request)
        assert result is None


def test_is_proxy_artifact_path(client):
    """Test _is_proxy_artifact_path function"""
    from mlflow_oidc_auth.hooks.before_request import _is_proxy_artifact_path

    # Test positive case
    assert _is_proxy_artifact_path("/api/2.0/mlflow-artifacts/artifacts/experiment1/file.txt") is True

    # Test negative case
    assert _is_proxy_artifact_path("/api/2.0/mlflow/experiments/search") is False


def test_get_proxy_artifact_validator_no_view_args(client):
    """Test _get_proxy_artifact_validator with no view_args (list operation)"""
    from mlflow_oidc_auth.hooks.before_request import _get_proxy_artifact_validator
    from mlflow_oidc_auth.validators import validate_can_read_experiment_artifact_proxy

    result = _get_proxy_artifact_validator("GET", None)
    assert result == validate_can_read_experiment_artifact_proxy


def test_get_proxy_artifact_validator_with_view_args(client):
    """Test _get_proxy_artifact_validator with view_args for different methods"""
    from mlflow_oidc_auth.hooks.before_request import _get_proxy_artifact_validator
    from mlflow_oidc_auth.validators import (
        validate_can_read_experiment_artifact_proxy,
        validate_can_update_experiment_artifact_proxy,
        validate_can_delete_experiment_artifact_proxy,
    )

    view_args = {"experiment_id": "123"}

    # Test GET (download)
    result = _get_proxy_artifact_validator("GET", view_args)
    assert result == validate_can_read_experiment_artifact_proxy

    # Test PUT (upload)
    result = _get_proxy_artifact_validator("PUT", view_args)
    assert result == validate_can_update_experiment_artifact_proxy

    # Test DELETE
    result = _get_proxy_artifact_validator("DELETE", view_args)
    assert result == validate_can_delete_experiment_artifact_proxy

    # Test unsupported method
    result = _get_proxy_artifact_validator("PATCH", view_args)
    assert result is None


def test_proxy_artifact_authorization_success(client):
    """Test proxy artifact path authorization success"""
    with app.test_request_context(path="/api/2.0/mlflow-artifacts/artifacts/experiment1/file.txt", method="GET"):
        session["username"] = "user"
        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request._find_validator", return_value=None
        ), patch("mlflow_oidc_auth.hooks.before_request.validate_can_read_experiment_artifact_proxy", return_value=True):
            response = before_request_hook()
            assert response is None  # Authorization succeeded


def test_proxy_artifact_authorization_failure(client):
    """Test proxy artifact path authorization failure"""
    with app.test_request_context(path="/api/2.0/mlflow-artifacts/artifacts/experiment1/file.txt", method="GET"):
        session["username"] = "user"
        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request._find_validator", return_value=None
        ), patch("mlflow_oidc_auth.hooks.before_request.validate_can_read_experiment_artifact_proxy", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request.responses.make_forbidden_response", return_value=Response("Forbidden", status=403)
        ) as mock_forbidden:
            response = before_request_hook()
            assert response.status_code == 403  # type: ignore
            mock_forbidden.assert_called_once()


def test_proxy_artifact_no_validator(client):
    """Test proxy artifact path when no validator is found"""
    with app.test_request_context(path="/api/2.0/mlflow-artifacts/artifacts/experiment1/file.txt", method="PATCH"):  # Unsupported method
        session["username"] = "user"
        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request._find_validator", return_value=None
        ):
            response = before_request_hook()
            assert response is None  # No validator, so no authorization check


def test_logged_model_route_authorization(client):
    """Test authorization for logged model routes"""
    with app.test_request_context(path="/api/2.0/mlflow/logged-models/12345", method="GET"):
        session["username"] = "user"
        mock_validator = MagicMock(return_value=True)

        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request._find_validator", return_value=mock_validator
        ):
            response = before_request_hook()
            assert response is None  # Authorization succeeded
            mock_validator.assert_called_once()


def test_logged_model_route_authorization_failure(client):
    """Test authorization failure for logged model routes"""
    with app.test_request_context(path="/api/2.0/mlflow/logged-models/12345", method="GET"):
        session["username"] = "user"
        mock_validator = MagicMock(return_value=False)

        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request._find_validator", return_value=mock_validator
        ), patch("mlflow_oidc_auth.hooks.before_request.responses.make_forbidden_response", return_value=Response("Forbidden", status=403)) as mock_forbidden:
            response = before_request_hook()
            assert response.status_code == 403  # type: ignore
            mock_validator.assert_called_once()
            mock_forbidden.assert_called_once()


def test_no_validator_found(client):
    """Test when no validator is found for a route"""
    with app.test_request_context(path="/unknown/route", method="GET"):
        session["username"] = "user"

        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request._find_validator", return_value=None
        ), patch("mlflow_oidc_auth.hooks.before_request._is_proxy_artifact_path", return_value=False):
            response = before_request_hook()
            assert response is None  # No validator, so no authorization check



import importlib
from unittest.mock import MagicMock, patch

import pytest

from mlflow_oidc_auth.auth import (
    _get_oidc_jwks,
    authenticate_request_basic_auth,
    authenticate_request_bearer_token,
    get_oauth_instance,
    validate_token,
)


class TestAuth:
    @patch("mlflow_oidc_auth.auth.OAuth")
    @patch("mlflow_oidc_auth.auth.config")
    def test_get_oauth_instance(self, mock_config, mock_oauth):
        mock_app = MagicMock()
        mock_oauth_instance = MagicMock()
        mock_oauth.return_value = mock_oauth_instance

        mock_config.OIDC_CLIENT_ID = "client_id"
        mock_config.OIDC_CLIENT_SECRET = "client_secret"
        mock_config.OIDC_DISCOVERY_URL = "discovery_url"
        mock_config.OIDC_SCOPE = "scope"

        result = get_oauth_instance(mock_app)

        mock_oauth.assert_called_once_with(mock_app)
        mock_oauth_instance.register.assert_called_once_with(
            name="oidc",
            client_id="client_id",
            client_secret="client_secret",
            server_metadata_url="discovery_url",
            client_kwargs={"scope": "scope"},
        )
        assert result == mock_oauth_instance

    @patch("mlflow_oidc_auth.auth.requests")
    @patch("mlflow_oidc_auth.auth.config")
    def test_get_oidc_jwks_success(self, mock_config, mock_requests):
        mock_cache = MagicMock()
        mock_app = MagicMock()
        mock_requests.get.return_value.json.return_value = {"jwks_uri": "jwks_uri"}
        mock_cache.get.return_value = None
        mock_config.OIDC_DISCOVERY_URL = "discovery_url"

        mlflow_oidc_app = importlib.import_module("mlflow_oidc_auth.app")
        with patch.object(mlflow_oidc_app, "cache", mock_cache), patch.object(mlflow_oidc_app, "app", mock_app):
            result = _get_oidc_jwks()
            mock_cache.set.assert_called_once_with("jwks", mock_requests.get.return_value.json.return_value, timeout=3600)
            assert result == mock_requests.get.return_value.json.return_value

    @patch("mlflow_oidc_auth.auth.app")
    def test_get_oidc_jwks_cache_hit(self, mock_app):
        mock_cache = MagicMock()
        mock_cache.get.return_value = {"keys": "cached_keys"}

        mlflow_oidc_app = importlib.import_module("mlflow_oidc_auth.app")
        with patch.object(mlflow_oidc_app, "cache", mock_cache):
            result = _get_oidc_jwks()
            assert result == {"keys": "cached_keys"}

    @patch("mlflow_oidc_auth.auth.config")
    def test_get_oidc_jwks_no_discovery_url(self, mock_config):
        mock_config.OIDC_DISCOVERY_URL = None
        mlflow_oidc_app = importlib.import_module("mlflow_oidc_auth.app")
        mock_cache = MagicMock()
        mock_cache.get.return_value = None

        with patch.object(mlflow_oidc_app, "cache", mock_cache):
            with pytest.raises(ValueError, match="OIDC_DISCOVERY_URL is not set"):
                _get_oidc_jwks()

    @patch("mlflow_oidc_auth.auth.config")
    def test_get_oidc_jwks_clear_cache(self, mock_config):
        mock_cache = MagicMock()
        mock_app = MagicMock()
        mock_config.OIDC_DISCOVERY_URL = "discovery_url"

        mlflow_oidc_app = importlib.import_module("mlflow_oidc_auth.app")
        with patch.object(mlflow_oidc_app, "cache", mock_cache), patch.object(mlflow_oidc_app, "app", mock_app):
            with patch("mlflow_oidc_auth.auth.requests") as mock_requests:
                mock_requests.get.return_value.json.return_value = {"jwks_uri": "jwks_uri"}
                mock_cache.get.return_value = None

                _get_oidc_jwks(clear_cache=True)
                mock_cache.delete.assert_called_once_with("jwks")

    @patch("mlflow_oidc_auth.auth._get_oidc_jwks")
    @patch("mlflow_oidc_auth.auth.jwt.decode")
    def test_validate_token_success(self, mock_jwt_decode, mock_get_oidc_jwks):
        mock_jwks = {"keys": "jwks"}
        mock_get_oidc_jwks.return_value = mock_jwks
        mock_payload = MagicMock()
        mock_jwt_decode.return_value = mock_payload

        result = validate_token("token")

        mock_jwt_decode.assert_called_once_with("token", mock_jwks)
        mock_payload.validate.assert_called_once()
        assert result == mock_payload

    @patch("mlflow_oidc_auth.auth._get_oidc_jwks")
    @patch("mlflow_oidc_auth.auth.jwt.decode")
    def test_validate_token_bad_signature_then_success(self, mock_jwt_decode, mock_get_oidc_jwks):
        from authlib.jose.errors import BadSignatureError

        mock_get_oidc_jwks.side_effect = [{"keys": "jwks1"}, {"keys": "jwks2"}]
        mock_payload = MagicMock()
        mock_jwt_decode.side_effect = [BadSignatureError("bad sig"), mock_payload]

        mlflow_oidc_app = importlib.import_module("mlflow_oidc_auth.app")
        with patch.object(mlflow_oidc_app, "app", MagicMock()):
            result = validate_token("token")
            assert result == mock_payload
            assert mock_get_oidc_jwks.call_count == 2

    @patch("mlflow_oidc_auth.auth._get_oidc_jwks")
    @patch("mlflow_oidc_auth.auth.jwt.decode")
    def test_validate_token_exception_after_refresh(self, mock_jwt_decode, mock_get_oidc_jwks):
        from authlib.jose.errors import BadSignatureError

        mock_get_oidc_jwks.side_effect = [{"keys": "jwks1"}, {"keys": "jwks2"}]
        mock_jwt_decode.side_effect = [BadSignatureError("bad sig"), Exception("other error")]

        mlflow_oidc_app = importlib.import_module("mlflow_oidc_auth.app")
        with patch.object(mlflow_oidc_app, "app", MagicMock()):
            with pytest.raises(Exception, match="other error"):
                validate_token("token")
            assert mock_get_oidc_jwks.call_count == 2

    @patch("mlflow_oidc_auth.auth.store")
    def test_authenticate_request_basic_auth_success(self, mock_store):
        mock_request = MagicMock()
        mock_request.authorization.username = "user"
        mock_request.authorization.password = "pass"
        mock_store.authenticate_user.return_value = True

        with patch("mlflow_oidc_auth.auth.request", mock_request):
            result = authenticate_request_basic_auth()
            mock_store.authenticate_user.assert_called_once_with("user", "pass")
            assert result is True

    def test_authenticate_request_basic_auth_no_auth(self):
        mock_request = MagicMock()
        mock_request.authorization = None

        with patch("mlflow_oidc_auth.auth.request", mock_request):
            assert authenticate_request_basic_auth() is False

    @patch("mlflow_oidc_auth.auth.store")
    def test_authenticate_request_basic_auth_invalid_credentials(self, mock_store):
        mock_request = MagicMock()
        mock_request.authorization.username = "user"
        mock_request.authorization.password = "wrong"
        mock_store.authenticate_user.return_value = False

        with patch("mlflow_oidc_auth.auth.request", mock_request), patch("mlflow_oidc_auth.auth.app"):
            assert authenticate_request_basic_auth() is False

    @patch("mlflow_oidc_auth.auth.validate_token")
    def test_authenticate_request_bearer_token_success(self, mock_validate_token):
        mock_request = MagicMock()
        mock_request.authorization.token = "token"
        mock_validate_token.return_value = {"email": "user@example.com"}

        with patch("mlflow_oidc_auth.auth.request", mock_request), patch("mlflow_oidc_auth.auth.app"):
            result = authenticate_request_bearer_token()
            mock_validate_token.assert_called_once_with("token")
            assert result is True

    def test_authenticate_request_bearer_token_no_auth(self):
        mock_request = MagicMock()
        mock_request.authorization = None

        with patch("mlflow_oidc_auth.auth.request", mock_request), patch("mlflow_oidc_auth.auth.app"):
            assert authenticate_request_bearer_token() is False

    @patch("mlflow_oidc_auth.auth.validate_token")
    def test_authenticate_request_bearer_token_invalid(self, mock_validate_token):
        mock_request = MagicMock()
        mock_request.authorization.token = "invalid"
        mock_validate_token.side_effect = Exception("Invalid token")

        with patch("mlflow_oidc_auth.auth.request", mock_request), patch("mlflow_oidc_auth.auth.app"):
            assert authenticate_request_bearer_token() is False

    def test_handle_token_validation_success(self):
        from mlflow_oidc_auth.auth import handle_token_validation

        oauth_instance = MagicMock()
        token = {"access_token": "token"}
        oauth_instance.oidc.authorize_access_token.return_value = token

        with patch("mlflow_oidc_auth.auth.app"):
            result = handle_token_validation(oauth_instance)
            assert result == token

    def test_handle_token_validation_bad_signature_recovery(self):
        from mlflow_oidc_auth.auth import handle_token_validation
        from authlib.jose.errors import BadSignatureError

        oauth_instance = MagicMock()
        oauth_instance.oidc.authorize_access_token.side_effect = [BadSignatureError(result=None), {"access_token": "token"}]

        mlflow_oidc_app = importlib.import_module("mlflow_oidc_auth.app")
        with patch.object(mlflow_oidc_app, "app", MagicMock()):
            result = handle_token_validation(oauth_instance)
            assert result == {"access_token": "token"}

    def test_handle_token_validation_bad_signature_fails(self):
        from mlflow_oidc_auth.auth import handle_token_validation
        from authlib.jose.errors import BadSignatureError

        oauth_instance = MagicMock()
        oauth_instance.oidc.authorize_access_token.side_effect = [
            BadSignatureError(result=None),
            BadSignatureError(result=None),
        ]

        with patch("mlflow_oidc_auth.auth.app", MagicMock()):
            result = handle_token_validation(oauth_instance)
            assert result is None

    def test_handle_user_and_group_management_success(self):
        from mlflow_oidc_auth.auth import handle_user_and_group_management

        token = {
            "userinfo": {"email": "admin@example.com", "name": "Admin", "groups": ["admin"]},
            "access_token": "token",
        }

        config = importlib.import_module("mlflow_oidc_auth.config").config
        config.OIDC_GROUP_DETECTION_PLUGIN = None
        config.OIDC_GROUPS_ATTRIBUTE = "groups"
        config.OIDC_ADMIN_GROUP_NAME = "admin"
        config.OIDC_GROUP_NAME = ["users"]

        with patch("mlflow_oidc_auth.auth.create_user") as mock_create, patch("mlflow_oidc_auth.auth.populate_groups") as mock_populate, patch(
            "mlflow_oidc_auth.auth.update_user"
        ) as mock_update, patch("mlflow_oidc_auth.auth.app"):
            errors = handle_user_and_group_management(token)
            assert errors == []
            mock_create.assert_called_once()
            mock_populate.assert_called_once()
            mock_update.assert_called_once()

    def test_handle_user_and_group_management_missing_profile(self):
        from mlflow_oidc_auth.auth import handle_user_and_group_management

        token = {"userinfo": {}, "access_token": "token"}
        errors = handle_user_and_group_management(token)
        assert "No email provided" in str(errors)
        assert "No display name provided" in str(errors)

    def test_handle_userinfo_missing_field_email_but_has_preferred_username_success(self):
        from mlflow_oidc_auth.auth import handle_user_and_group_management

        token = {"userinfo": {"name": "Test Tes", "preferred_username": "techaccount@example.net", "groups": ["users"]}, "access_token": "token"}
        config = importlib.import_module("mlflow_oidc_auth.config").config
        config.OIDC_GROUP_DETECTION_PLUGIN = None
        config.OIDC_GROUPS_ATTRIBUTE = "groups"
        config.OIDC_ADMIN_GROUP_NAME = "admin"
        config.OIDC_GROUP_NAME = ["users"]

        with patch("mlflow_oidc_auth.auth.create_user") as mock_create, patch("mlflow_oidc_auth.auth.populate_groups") as mock_populate, patch(
            "mlflow_oidc_auth.auth.update_user"
        ) as mock_update, patch("mlflow_oidc_auth.auth.app"):
            errors = handle_user_and_group_management(token)
            assert errors == []
            mock_create.assert_called_once()
            mock_populate.assert_called_once()
            mock_update.assert_called_once()

    def test_handle_user_and_group_management_unauthorized(self):
        from mlflow_oidc_auth.auth import handle_user_and_group_management

        token = {"userinfo": {"email": "user@example.com", "name": "User", "groups": ["guests"]}}

        config = importlib.import_module("mlflow_oidc_auth.config").config
        config.OIDC_GROUP_DETECTION_PLUGIN = None
        config.OIDC_GROUPS_ATTRIBUTE = "groups"
        config.OIDC_ADMIN_GROUP_NAME = "admin"
        config.OIDC_GROUP_NAME = ["users"]

        with patch("mlflow_oidc_auth.auth.app"):
            errors = handle_user_and_group_management(token)
            assert "not allowed to login" in str(errors)

    def test_handle_user_and_group_management_group_plugin_error(self):
        from mlflow_oidc_auth.auth import handle_user_and_group_management

        token = {
            "userinfo": {"email": "user@example.com", "name": "User"},
            "access_token": "token",
        }

        config = importlib.import_module("mlflow_oidc_auth.config").config
        config.OIDC_GROUP_DETECTION_PLUGIN = "nonexistent.module"

        with patch("mlflow_oidc_auth.auth.app"):
            errors = handle_user_and_group_management(token)
            assert "Group detection error: Failed to get user groups" in errors

    def test_handle_user_and_group_management_group_missing_error(self):
        from mlflow_oidc_auth.auth import handle_user_and_group_management

        token = {
            "userinfo": {"email": "user@example.com", "name": "User"},
            "access_token": "token",
        }

        config = importlib.import_module("mlflow_oidc_auth.config").config
        config.OIDC_GROUP_DETECTION_PLUGIN = None
        config.OIDC_GROUPS_ATTRIBUTE = "groups"

        with patch("mlflow_oidc_auth.auth.app"):
            errors = handle_user_and_group_management(token)
            assert "Group detection error: Failed to get user groups" in errors

    def test_handle_user_and_group_management_db_error(self):
        from mlflow_oidc_auth.auth import handle_user_and_group_management

        token = {
            "userinfo": {"email": "admin@example.com", "name": "Admin", "groups": ["admin"]},
            "access_token": "token",
        }

        config = importlib.import_module("mlflow_oidc_auth.config").config
        config.OIDC_GROUP_DETECTION_PLUGIN = None
        config.OIDC_GROUPS_ATTRIBUTE = "groups"
        config.OIDC_ADMIN_GROUP_NAME = "admin"
        config.OIDC_GROUP_NAME = ["users"]

        with patch("mlflow_oidc_auth.auth.create_user", side_effect=Exception("DB error")), patch("mlflow_oidc_auth.auth.populate_groups"), patch(
            "mlflow_oidc_auth.auth.update_user"
        ), patch("mlflow_oidc_auth.auth.app"):
            errors = handle_user_and_group_management(token)
            assert "User/group DB error: Failed to update user/groups" in errors

    def test_process_oidc_callback_success(self):
        from mlflow_oidc_auth.auth import process_oidc_callback

        mock_request = MagicMock()
        mock_request.args.get.side_effect = lambda k: "state_value" if k == "state" else None
        session = {"oauth_state": "state_value"}
        token = {"userinfo": {"email": "user@example.com"}}

        with patch("mlflow_oidc_auth.auth.get_oauth_instance") as mock_oauth, patch("mlflow_oidc_auth.auth.handle_token_validation", return_value=token), patch(
            "mlflow_oidc_auth.auth.handle_user_and_group_management", return_value=[]
        ), patch("mlflow_oidc_auth.auth.app"):
            mock_oauth.return_value.oidc = MagicMock()
            email, errors = process_oidc_callback(mock_request, session)
            assert email == "user@example.com"
            assert errors == []

    def test_process_oidc_callback_oidc_error(self):
        from mlflow_oidc_auth.auth import process_oidc_callback

        mock_request = MagicMock()
        mock_request.args.get.side_effect = lambda k: "error" if k == "error" else "description"

        email, errors = process_oidc_callback(mock_request, {})
        assert email is None
        assert "OIDC provider error" in str(errors)

    def test_process_oidc_callback_state_mismatch(self):
        from mlflow_oidc_auth.auth import process_oidc_callback

        mock_request = MagicMock()
        mock_request.args.get.side_effect = lambda k: "wrong_state" if k == "state" else None
        session = {"oauth_state": "correct_state"}

        email, errors = process_oidc_callback(mock_request, session)
        assert email is None
        assert "Invalid state parameter" in str(errors)

    def test_process_oidc_callback_missing_oauth_state(self):
        from mlflow_oidc_auth.auth import process_oidc_callback

        mock_request = MagicMock()
        mock_request.args.get.side_effect = lambda k: "state_value" if k == "state" else None
        session = {}  # Missing oauth_state

        email, errors = process_oidc_callback(mock_request, session)
        assert email is None
        assert "Missing OAuth state in session" in str(errors)

    def test_process_oidc_callback_oauth_instance_none(self):
        from mlflow_oidc_auth.auth import process_oidc_callback

        mock_request = MagicMock()
        mock_request.args.get.side_effect = lambda k: "state_value" if k == "state" else None
        session = {"oauth_state": "state_value"}

        with patch("mlflow_oidc_auth.auth.get_oauth_instance", return_value=None), patch("mlflow_oidc_auth.auth.app"):
            email, errors = process_oidc_callback(mock_request, session)
            assert email is None
            assert "OAuth instance or OIDC is not properly initialized" in str(errors)

    def test_process_oidc_callback_oauth_instance_no_oidc(self):
        from mlflow_oidc_auth.auth import process_oidc_callback

        mock_request = MagicMock()
        mock_request.args.get.side_effect = lambda k: "state_value" if k == "state" else None
        session = {"oauth_state": "state_value"}

        class DummyOAuth:
            pass  # No oidc attribute

        with patch("mlflow_oidc_auth.auth.get_oauth_instance", return_value=DummyOAuth()), patch("mlflow_oidc_auth.auth.app"):
            email, errors = process_oidc_callback(mock_request, session)
            assert email is None
            assert "OAuth instance or OIDC is not properly initialized" in str(errors)

    def test_process_oidc_callback_token_validation_none(self):
        from mlflow_oidc_auth.auth import process_oidc_callback

        mock_request = MagicMock()
        mock_request.args.get.side_effect = lambda k: "state_value" if k == "state" else None
        session = {"oauth_state": "state_value"}

        with patch("mlflow_oidc_auth.auth.get_oauth_instance") as mock_oauth, patch("mlflow_oidc_auth.auth.handle_token_validation", return_value=None), patch(
            "mlflow_oidc_auth.auth.app"
        ):
            mock_oauth.return_value.oidc = MagicMock()
            email, errors = process_oidc_callback(mock_request, session)
            assert email is None
            assert "Invalid token signature or token could not be validated" in str(errors)

    def test_process_oidc_callback_user_management_errors(self):
        from mlflow_oidc_auth.auth import process_oidc_callback

        mock_request = MagicMock()
        mock_request.args.get.side_effect = lambda k: "state_value" if k == "state" else None
        session = {"oauth_state": "state_value"}
        token = {"userinfo": {"email": "user@example.com"}}

        with patch("mlflow_oidc_auth.auth.get_oauth_instance") as mock_oauth, patch("mlflow_oidc_auth.auth.handle_token_validation", return_value=token), patch(
            "mlflow_oidc_auth.auth.handle_user_and_group_management", return_value=["Some error"]
        ), patch("mlflow_oidc_auth.auth.app"):
            mock_oauth.return_value.oidc = MagicMock()
            email, errors = process_oidc_callback(mock_request, session)
            assert email is None
            assert "Some error" in errors



import os

import redis

SESSION_TYPE = "redis"
SESSION_REDIS = redis.Redis(
    host=os.environ.get("REDIS_HOST", "localhost"),
    port=int(os.environ.get("REDIS_PORT", 6379)),
    db=int(os.environ.get("REDIS_DB", 0)),
    password=os.environ.get("REDIS_PASSWORD", None),
    ssl=os.environ.get("REDIS_SSL", str(False)).lower() in ("true", "1", "t"),
    username=os.environ.get("REDIS_USERNAME", None),
)
